#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import os
from collections.abc import Iterable
from typing import Union

import numpy as np
import onnxruntime as ort
from PIL import Image

MODEL_CHECKPOINT_FN = "model.pt"
DEFAULT_NUM_THREADS = 4


# Classes:
CLASSES_MAP = {
    0: "background",
    1: "Caption",
    2: "Footnote",
    3: "Formula",
    4: "List-item",
    5: "Page-footer",
    6: "Page-header",
    7: "Picture",
    8: "Section-header",
    9: "Table",
    10: "Text",
    11: "Title",
    12: "Document Index",
    13: "Code",
    14: "Checkbox-Selected",
    15: "Checkbox-Unselected",
    16: "Form",
    17: "Key-Value Region",
}


class LayoutPredictor:
    r"""
    Document layout prediction using ONNX
    """

    def __init__(
        self, artifact_path: str, num_threads: int = None, use_cpu_only: bool = False
    ):
        r"""
        Provide the artifact path that contains the LayoutModel ONNX file

        The number of threads is decided, in the following order, by:
        1. The init method parameter `num_threads`, if it is set.
        2. The envvar "OMP_NUM_THREADS", if it is set.
        3. The default value DEFAULT_NUM_THREADS.

        The execution provided is decided, in the following order:
        1. If the init method parameter `cpu_only` is True or the envvar "USE_CPU_ONLY" is set,
           it uses the "CPUExecutionProvider".
        3. Otherwise if the "CUDAExecutionProvider" is present, use:
            ["CUDAExecutionProvider", "CPUExecutionProvider"]:

        Parameters
        ----------
        artifact_path: Path for the model ONNX file.
        num_threads: (Optional) Number of threads to run the inference.
        use_cpu_only: (Optional) If True, it forces CPU as the execution provider.

        Raises
        ------
        FileNotFoundError when the model's ONNX file is missing
        """
        # Set basic params
        self._threshold = 0.6  # Score threshold
        self._image_size = 640
        self._size = np.asarray([[self._image_size, self._image_size]], dtype=np.int64)

        # Get env vars
        self._use_cpu_only = use_cpu_only or ("USE_CPU_ONLY" in os.environ)
        if num_threads is None:
            num_threads = int(os.environ.get("OMP_NUM_THREADS", DEFAULT_NUM_THREADS))
        self._num_threads = num_threads

        # Decide the execution providers
        if (
            not self._use_cpu_only
            and "CUDAExecutionProvider" in ort.get_available_providers()
        ):
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
        else:
            providers = ["CPUExecutionProvider"]
        self._providers = providers

        # Model ONNX file
        self._onnx_fn = os.path.join(artifact_path, MODEL_CHECKPOINT_FN)
        if not os.path.isfile(self._onnx_fn):
            raise FileNotFoundError("Missing ONNX file: {}".format(self._onnx_fn))

        # ONNX options
        self._options = ort.SessionOptions()
        self._options.intra_op_num_threads = self._num_threads
        self.sess = ort.InferenceSession(
            self._onnx_fn,
            sess_options=self._options,
            providers=self._providers,
        )

    def info(self) -> dict:
        r"""
        Get information about the configuration of LayoutPredictor
        """
        info = {
            "onnx_file": self._onnx_fn,
            "intra_op_num_threads": self._num_threads,
            "use_cpu_only": self._use_cpu_only,
            "providers": self._providers,
            "image_size": self._image_size,
            "threshold": self._threshold,
        }
        return info

    def predict(self, orig_img: Union[Image.Image, np.ndarray]) -> Iterable[dict]:
        r"""
        Predict bounding boxes for a given image.
        The origin (0, 0) is the top-left corner and the predicted bbox coords are provided as:
        [left, top, right, bottom]

        Parameter
        ---------
        origin_img: Image to be predicted as a PIL Image object or numpy array.

        Yield
        -----
        Bounding box as a dict with the keys: "label", "confidence", "l", "t", "r", "b"

        Raises
        ------
        TypeError when the input image is not supported
        """
        # Convert image format
        if isinstance(orig_img, Image.Image):
            page_img = orig_img.convert("RGB")
        elif isinstance(orig_img, np.ndarray):
            page_img = Image.fromarray(orig_img).convert("RGB")
        else:
            raise TypeError("Not supported input image format")

        w, h = page_img.size
        page_img = page_img.resize((self._image_size, self._image_size))
        page_data = np.array(page_img, dtype=np.uint8) / np.float32(255.0)
        page_data = np.expand_dims(np.transpose(page_data, axes=[2, 0, 1]), axis=0)

        # Predict
        labels, boxes, scores = self.sess.run(
            output_names=None,
            input_feed={
                "images": page_data,
                "orig_target_sizes": self._size,
            },
        )

        # Yield output
        for label, box, score in zip(labels[0], boxes[0], scores[0]):
            if score > self._threshold:
                yield {
                    "l": box[0] / self._image_size * w,
                    "t": box[1] / self._image_size * h,
                    "r": box[2] / self._image_size * w,
                    "b": box[3] / self._image_size * h,
                    "label": CLASSES_MAP[label],
                    "confidence": score,
                }



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import torch


def model_info(model, verbose=False):
    # Plots a line-by-line description of a PyTorch model
    n_p = sum(x.numel() for x in model.parameters())  # number parameters
    n_g = sum(
        x.numel() for x in model.parameters() if x.requires_grad
    )  # number gradients
    if verbose:
        print(
            "%5s %40s %9s %12s %20s %10s %10s"
            % ("layer", "name", "gradient", "parameters", "shape", "mu", "sigma")
        )
        for i, (name, p) in enumerate(model.named_parameters()):
            name = name.replace("module_list.", "")
            print(
                "%5g %40s %9s %12g %20s %10.3g %10.3g"
                % (
                    i,
                    name,
                    p.requires_grad,
                    p.numel(),
                    list(p.shape),
                    p.mean(),
                    p.std(),
                )
            )

    try:  # FLOPS
        from thop import profile

        macs, _ = profile(model, inputs=(torch.zeros(1, 3, 480, 640),), verbose=False)
        fs = ", %.1f GFLOPS" % (macs / 1e9 * 2)
    except Exception:
        fs = ""

    print(
        "Model Summary: %g layers, %g parameters, %g gradients%s"
        % (len(list(model.parameters())), n_p, n_g, fs)
    )


# def init_seeds(seed=0):
#     torch.manual_seed(seed)
#
#     # Reduce randomness (may be slower on Tesla GPUs)
#     # https://pytorch.org/docs/stable/notes/randomness.html
#     if seed == 0:
#         cudnn.deterministic = False
#         cudnn.benchmark = True
#
#
# def select_device(device='', apex=False, batch_size=None):
#     # device = 'cpu' or '0' or '0,1,2,3'
#     cpu_request = device.lower() == 'cpu'
#     if device and not cpu_request:  # if device requested other than 'cpu'
#         os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
#         # check availablity
#         assert torch.cuda.is_available(), 'CUDA unavailable, invalid device %s requested' % device
#
#     cuda = False if cpu_request else torch.cuda.is_available()
#     if cuda:
#         c = 1024 ** 2  # bytes to MB
#         ng = torch.cuda.device_count()
#         if ng > 1 and batch_size:  # check that batch_size is compatible with device_count
#             assert batch_size % ng == 0, 'batch-size %g not multiple of GPU count %g' % \
#                 (batch_size, ng)
#         x = [torch.cuda.get_device_properties(i) for i in range(ng)]
#         # apex for mixed precision https://github.com/NVIDIA/apex
#         s = 'Using CUDA ' + ('Apex ' if apex else '')
#         for i in range(0, ng):
#             if i == 1:
#                 s = ' ' * len(s)
#             print("%sdevice%g _CudaDeviceProperties(name='%s', total_memory=%dMB)" %
#                   (s, i, x[i].name, x[i].total_memory / c))
#     else:
#         print('Using CPU')
#
#     print('')  # skip a line
#     return torch.device('cuda:0' if cuda else 'cpu')
#
#
# def time_synchronized():
#     torch.cuda.synchronize() if torch.cuda.is_available() else None
#     return time.time()
#
#
# def initialize_weights(model):
#     for m in model.modules():
#         t = type(m)
#         if t is nn.Conv2d:
#             pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
#         elif t is nn.BatchNorm2d:
#             m.eps = 1e-4
#             m.momentum = 0.03
#         elif t in [nn.LeakyReLU, nn.ReLU, nn.ReLU6]:
#             m.inplace = True
#
#
# def find_modules(model, mclass=nn.Conv2d):
#     # finds layer indices matching module class 'mclass'
#     return [i for i, m in enumerate(model.module_list) if isinstance(m, mclass)]
#
#
# def fuse_conv_and_bn(conv, bn):
#     # https://tehnokv.com/posts/fusing-batchnorm-and-conv/
#     with torch.no_grad():
#         # init
#         fusedconv = torch.nn.Conv2d(conv.in_channels,
#                                     conv.out_channels,
#                                     kernel_size=conv.kernel_size,
#                                     stride=conv.stride,
#                                     padding=conv.padding,
#                                     bias=True)
#
#         # prepare filters
#         w_conv = conv.weight.clone().view(conv.out_channels, -1)
#         w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
#         fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.size()))
#
#         # prepare spatial bias
#         if conv.bias is not None:
#             b_conv = conv.bias
#         else:
#             b_conv = torch.zeros(conv.weight.size(0))
#         b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
#         fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)
#
#         return fusedconv
#
#
# def load_classifier(name='resnet101', n=2):
#     # Loads a pretrained model reshaped to n-class output
#     import pretrainedmodels  # https://github.com/Cadene/pretrained-models.pytorch#torchvision
#     model = pretrainedmodels.__dict__[name](num_classes=1000, pretrained='imagenet')
#
#     # Display model properties
#     for x in ['model.input_size', 'model.input_space', 'model.input_range', 'model.mean',
#               'model.std']:
#         print(x + ' =', eval(x))
#
#     # Reshape output to n classes
#     filters = model.last_linear.weight.shape[1]
#     model.last_linear.bias = torch.nn.Parameter(torch.zeros(n))
#     model.last_linear.weight = torch.nn.Parameter(torch.zeros(n, filters))
#     model.last_linear.out_features = n
#     return model
#
#
# def scale_img(img, ratio=1.0, same_shape=True):  # img(16,3,256,416), r=ratio
#     # scales img(bs,3,y,x) by ratio
#     h, w = img.shape[2:]
#     s = (int(h * ratio), int(w * ratio))  # new size
#     img = F.interpolate(img, size=s, mode='bilinear', align_corners=False)  # resize
#     if not same_shape:  # pad/crop img
#         gs = 64  # (pixels) grid size
#         h, w = [math.ceil(x * ratio / gs) * gs for x in (h, w)]
#     return F.pad(img, [0, w - s[1], 0, h - s[0]], value=0.447)  # value = imagenet mean
#
#
# class ModelEMA:
#     """ Model Exponential Moving Average from https://github.com/rwightman/pytorch-image-models
#     Keep a moving average of everything in the model state_dict (parameters and buffers).
#     This is intended to allow functionality like
#     https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
#     A smoothed version of the weights is necessary for some training schemes to perform well.
#     E.g. Google's hyper-params for training MNASNet, MobileNet-V3, EfficientNet, etc that use
#     RMSprop with a short 2.4-3 epoch decay period and slow LR decay rate of .96-.99 requires EMA
#     smoothing of weights to match results. Pay attention to the decay constant you are using
#     relative to your update count per epoch.
#     To keep EMA from using GPU resources, set device='cpu'. This will save a bit of memory but
#     disable validation of the EMA weights. Validation will have to be done manually in a separate
#     process, or after the training stops converging.
#     This class is sensitive where it is initialized in the sequence of model init,
#     GPU assignment and distributed training wrappers.
#     I've tested with the sequence in my own train.py for torch.DataParallel, apex.DDP, and
#     single-GPU.
#     """
#
#     def __init__(self, model, decay=0.9999, device=''):
#         # make a copy of the model for accumulating moving average of weights
#         self.ema = deepcopy(model)
#         self.ema.eval()
#         self.updates = 0  # number of EMA updates
#         # decay exponential ramp (to help early epochs)
#         self.decay = lambda x: decay * (1 - math.exp(-x / 2000))
#         self.device = device  # perform ema on different device from model if set
#         if device:
#             self.ema.to(device=device)
#         for p in self.ema.parameters():
#             p.requires_grad_(False)
#
#     def update(self, model):
#         self.updates += 1
#         d = self.decay(self.updates)
#         with torch.no_grad():
#             if type(model) in (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel):
#                 msd, esd = model.module.state_dict(), self.ema.module.state_dict()
#             else:
#                 msd, esd = model.state_dict(), self.ema.state_dict()
#
#             for k, v in esd.items():
#                 if v.dtype.is_floating_point:
#                     v *= d
#                     v += (1. - d) * msd[k].detach()
#
#     def update_attr(self, model):
#         # Assign attributes (which may change during training)
#         for k in model.__dict__.keys():
#             if not k.startswith('_'):
#                 setattr(self.ema, k, getattr(model, k))



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging

import docling_ibm_models.tableformer.common as c
from docling_ibm_models.tableformer.data_management.tf_dataset import TFDataset

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


def dataset_test(config):
    r"""
    Parameters
    ----------
    config : dictionary
        The configuration settings
    """

    # model_type = config["model"]["type"]
    # Create the device and the Dataset
    device = "cpu"
    dataset = TFDataset(config, "train")
    dataset.set_device(device)

    # Loop over the data
    dataset.reset()
    dataset.shuffle()
    for i, batch in enumerate(dataset):
        print("Loading batch: {}".format(i))


if __name__ == "__main__":
    config = c.parse_arguments()
    dataset_test(config)



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import numbers
from collections.abc import Iterable, Sequence

import cv2
import numpy as np
import torch
from torchvision.transforms import functional

cv2.setNumThreads(0)
cv2.ocl.setUseOpenCL(False)

INTER_MODE = {
    "NEAREST": cv2.INTER_NEAREST,
    "BILINEAR": cv2.INTER_LINEAR,
    "BICUBIC": cv2.INTER_CUBIC,
}

PAD_MOD = {
    "constant": cv2.BORDER_CONSTANT,
    "edge": cv2.BORDER_REPLICATE,
    "reflect": cv2.BORDER_DEFAULT,
    "symmetric": cv2.BORDER_REFLECT,
}


def _is_tensor_image(img):
    return torch.is_tensor(img) and img.ndimension() == 3


def _is_numpy_image(img):
    return isinstance(img, np.ndarray) and (img.ndim in {2, 3})


def to_tensor(pic):
    """Converts a numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    Args:
        pic (np.ndarray, torch.Tensor): Image to be converted to tensor, (H x W x C[RGB]).
    Returns:
        Tensor: Converted image.
    """

    if _is_numpy_image(pic):
        if len(pic.shape) == 2:
            pic = cv2.cvtColor(pic, cv2.COLOR_GRAY2RGB)
        img = torch.from_numpy(pic.transpose((2, 0, 1)))
        # backward compatibility
        if isinstance(img, torch.ByteTensor) or img.max() > 1:
            return img.float().div(255)
        else:
            return img
    elif _is_tensor_image(pic):
        return pic

    else:
        try:
            return to_tensor(np.array(pic))
        except Exception:
            raise TypeError("pic should be ndarray. Got {}".format(type(pic)))


def to_cv_image(pic, mode=None):
    """Convert a tensor to an ndarray.
    Args:
        pic (Tensor or numpy.ndarray): Image to be converted to PIL Image.
        mode (str): color space and pixel depth of input data (optional)
                    for example: cv2.COLOR_RGB2BGR.
    Returns:
        np.array: Image converted to PIL Image.
    """
    if not (_is_numpy_image(pic) or _is_tensor_image(pic)):
        raise TypeError("pic should be Tensor or ndarray. Got {}.".format(type(pic)))

    npimg = pic
    if isinstance(pic, torch.FloatTensor):
        pic = pic.mul(255).byte()
    if torch.is_tensor(pic):
        npimg = np.squeeze(np.transpose(pic.numpy(), (1, 2, 0)))

    if not isinstance(npimg, np.ndarray):
        raise TypeError("Input pic must be a torch.Tensor or NumPy ndarray")
    if mode is None:
        return npimg

    else:
        return cv2.cvtColor(npimg, mode)


def normalize(tensor, mean, std):
    """Normalize a tensor image with mean and standard deviation.
    See ``Normalize`` for more details.
    Args:
        tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channely.
    Returns:
        Tensor: Normalized Tensor image.
    """
    if _is_tensor_image(tensor):
        for t, m, s in zip(tensor, mean, std, strict=False):
            t.sub_(m).div_(s)
        return tensor
    elif _is_numpy_image(tensor):
        return (tensor.astype(np.float32) - 255.0 * np.array(mean)) / np.array(std)
    else:
        raise RuntimeError("Undefined type")


def resize(img, size, interpolation="BILINEAR"):
    """Resize the input CV Image to the given size.
    Args:
        img (np.ndarray): Image to be resized.
        size (tuple or int): Desired output size. If size is a sequence like
            (h, w), the output size will be matched to this. If size is an int,
            the smaller edge of the image will be matched to this number maintaing
            the aspect ratio. i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (str, optional): Desired interpolation. Default is ``BILINEAR``
    Returns:
        cv Image: Resized image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))
    if not (isinstance(size, int) or (isinstance(size, Iterable) and len(size) == 2)):
        raise TypeError("Got inappropriate size arg: {}".format(size))

    if isinstance(size, int):
        h, w, c = img.shape
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return cv2.resize(
                img, dsize=(ow, oh), interpolation=INTER_MODE[interpolation]
            )
        else:
            oh = size
            ow = int(size * w / h)
            return cv2.resize(
                img, dsize=(ow, oh), interpolation=INTER_MODE[interpolation]
            )
    else:
        oh, ow = size
        return cv2.resize(
            img, dsize=(int(ow), int(oh)), interpolation=INTER_MODE[interpolation]
        )


def to_rgb_bgr(pic):
    """Converts a color image stored in BGR sequence to RGB (BGR to RGB)
    or stored in RGB sequence to BGR (RGB to BGR).
    Args:
        pic (np.ndarray, torch.Tensor): Image to be converted, (H x W x 3).
    Returns:
        Tensor: Converted image.
    """

    if _is_numpy_image(pic) or _is_tensor_image(pic):
        img = pic[:, :, [2, 1, 0]]
        return img
    else:
        try:
            return to_rgb_bgr(np.array(pic))
        except Exception:
            raise TypeError("pic should be numpy.ndarray or torch.Tensor.")


def pad(img, padding, fill=(0, 0, 0), padding_mode="constant"):
    """Pad the given CV Image on all sides with speficified padding mode and fill value.
    Args:
        img (np.ndarray): Image to be padded.
        padding (int or tuple): Padding on each border. If a single int is provided this
            is used to pad all borders. If tuple of length 2 is provided this is the padding
            on left/right and top/bottom respectively. If a tuple of length 4 is provided
            this is the padding for the left, top, right and bottom borders
            respectively.
        fill (int, tuple): Pixel fill value for constant fill. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.
            This value is only used when the padding_mode is constant
        padding_mode: Type of padding. Should be: constant, edge, reflect or symmetric.
        Default is constant.
            constant: pads with a constant value, this value is specified with fill
            edge: pads with the last value on the edge of the image
            reflect: pads with reflection of image (without repeating the last value on the edge)
                padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                will result in [3, 2, 1, 2, 3, 4, 3, 2]
            symmetric: pads with reflection of image (repeating the last value on the edge)
                padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                will result in [2, 1, 1, 2, 3, 4, 4, 3]
    Returns:
        CV Image: Padded image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    if not isinstance(padding, (numbers.Number, tuple)):
        raise TypeError("Got inappropriate padding arg")
    if not isinstance(fill, (numbers.Number, str, tuple)):
        raise TypeError("Got inappropriate fill arg")
    if not isinstance(padding_mode, str):
        raise TypeError("Got inappropriate padding_mode arg")

    if isinstance(padding, Sequence) and len(padding) not in [2, 4]:
        raise ValueError("Padding must be an int or a 2, or 4 element tuple")

    assert padding_mode in [
        "constant",
        "edge",
        "reflect",
        "symmetric",
    ], "Padding mode should be either constant, edge, reflect or symmetric"

    if isinstance(padding, int):
        pad_left = pad_right = pad_top = pad_bottom = padding
    if isinstance(padding, Sequence) and len(padding) == 2:
        pad_left = pad_right = padding[0]
        pad_top = pad_bottom = padding[1]
    if isinstance(padding, Sequence) and len(padding) == 4:
        pad_left, pad_top, pad_right, pad_bottom = padding

    if isinstance(fill, numbers.Number):
        fill = (fill,) * (2 * len(img.shape) - 3)

    if padding_mode == "constant":
        assert (len(fill) == 3 and len(img.shape) == 3) or (
            len(fill) == 1 and len(img.shape) == 2
        ), "channel of image is {} but length of fill is {}".format(
            img.shape[-1], len(fill)
        )

    img = cv2.copyMakeBorder(
        src=img,
        top=pad_top,
        bottom=pad_bottom,
        left=pad_left,
        right=pad_right,
        borderType=PAD_MOD[padding_mode],
        value=fill,
    )
    return img


def crop(img, x, y, h, w):
    """Crop the given CV Image.
    Args:
        img (np.ndarray): Image to be cropped.
        x: Upper pixel coordinate.
        y: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.
    Returns:
        CV Image: Cropped image.
    """
    assert _is_numpy_image(img), "img should be CV Image. Got {}".format(type(img))
    assert h > 0 and w > 0, "h={} and w={} should greater than 0".format(h, w)

    x1, y1, x2, y2 = round(x), round(y), round(x + h), round(y + w)

    #     try:
    #         check_point1 = img[x1, y1, ...]
    #         check_point2 = img[x2-1, y2-1, ...]
    #     except IndexError:
    #         img = cv2.copyMakeBorder(img, - min(0, x1), max(x2 - img.shape[0], 0),
    #                                  -min(0, y1), max(y2 - img.shape[1], 0),
    #                                  cv2.BORDER_CONSTANT, value=[0, 0, 0])
    #         y2 += -min(0, y1)
    #         y1 += -min(0, y1)
    #         x2 += -min(0, x1)
    #         x1 += -min(0, x1)
    #
    #     finally:
    #         return img[x1:x2, y1:y2, ...].copy()
    return img[x1:x2, y1:y2, ...].copy()


def center_crop(img, output_size):
    if isinstance(output_size, numbers.Number):
        output_size = (int(output_size), int(output_size))
    h, w, _ = img.shape
    th, tw = output_size
    i = int(round((h - th) * 0.5))
    j = int(round((w - tw) * 0.5))
    return crop(img, i, j, th, tw)


def resized_crop(img, i, j, h, w, size, interpolation="BILINEAR"):
    """Crop the given CV Image and resize it to desired size. Notably used in RandomResizedCrop.
    Args:
        img (np.ndarray): Image to be cropped.
        i: Upper pixel coordinate.
        j: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.
        size (sequence or int): Desired output size. Same semantics as ``scale``.
        interpolation (str, optional): Desired interpolation. Default is
            ``BILINEAR``.
    Returns:
        np.ndarray: Cropped image.
    """
    assert _is_numpy_image(img), "img should be CV Image"
    img = crop(img, i, j, h, w)
    img = resize(img, size, interpolation)
    return img


def hflip(img):
    """Horizontally flip the given PIL Image.
    Args:
        img (np.ndarray): Image to be flipped.
    Returns:
        np.ndarray:  Horizontall flipped image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    return cv2.flip(img, 1)


def vflip(img):
    """Vertically flip the given PIL Image.
    Args:
        img (CV Image): Image to be flipped.
    Returns:
        PIL Image:  Vertically flipped image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be PIL Image. Got {}".format(type(img)))

    return cv2.flip(img, 0)


def five_crop(img, size):
    """Crop the given CV Image into four corners and the central crop.
    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.
    Args:
       size (sequence or int): Desired output size of the crop. If size is an
           int instead of sequence like (h, w), a square crop (size, size) is
           made.
    Returns:
        tuple: tuple (tl, tr, bl, br, center) corresponding top left,
            top right, bottom left, bottom right and center crop.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    h, w, _ = img.shape
    crop_h, crop_w = size
    if crop_w > w or crop_h > h:
        raise ValueError(
            "Requested crop size {} is bigger than input size {}".format(size, (h, w))
        )
    tl = crop(img, 0, 0, crop_h, crop_w)
    tr = crop(img, 0, w - crop_w, crop_h, crop_w)
    bl = crop(img, h - crop_h, 0, crop_h, crop_w)
    br = crop(img, h - crop_h, w - crop_w, crop_h, crop_w)
    center = center_crop(img, (crop_h, crop_w))
    return (tl, tr, bl, br, center)


def ten_crop(img, size, vertical_flip=False):
    """Crop the given CV Image into four corners and the central crop plus the
       flipped version of these (horizontal flipping is used by default).
    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.
       Args:
           size (sequence or int): Desired output size of the crop. If size is an
               int instead of sequence like (h, w), a square crop (size, size) is
               made.
           vertical_flip (bool): Use vertical flipping instead of horizontal
        Returns:
            tuple: tuple (tl, tr, bl, br, center, tl_flip, tr_flip, bl_flip,
                br_flip, center_flip) corresponding top left, top right,
                bottom left, bottom right and center crop and same for the
                flipped image.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    first_five = five_crop(img, size)

    if vertical_flip:
        img = vflip(img)
    else:
        img = hflip(img)

    second_five = five_crop(img, size)
    return first_five + second_five


def adjust_brightness(img, brightness_factor):
    """Adjust brightness of an Image.
    Args:
        img (np.ndarray): CV Image to be adjusted.
        brightness_factor (float):  How much to adjust the brightness. Can be
            any non negative number. 0 gives a black image, 1 gives the
            original image while 2 increases the brightness by a factor of 2.
    Returns:
        np.ndarray: Brightness adjusted image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    im = img.astype(np.float32) * brightness_factor
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_contrast(img, contrast_factor):
    """Adjust contrast of an Image.
    Args:
        img (np.ndarray): CV Image to be adjusted.
        contrast_factor (float): How much to adjust the contrast. Can be any
            non negative number. 0 gives a solid gray image, 1 gives the
            original image while 2 increases the contrast by a factor of 2.
    Returns:
        np.ndarray: Contrast adjusted image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))
    im = img.astype(np.float32)
    mean = round(cv2.cvtColor(im, cv2.COLOR_RGB2GRAY).mean())
    im = (1 - contrast_factor) * mean + contrast_factor * im
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_saturation(img, saturation_factor):
    """Adjust color saturation of an image.
    Args:
        img (np.ndarray): CV Image to be adjusted.
        saturation_factor (float):  How much to adjust the saturation. 0 will
            give a gray image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.
    Returns:
        np.ndarray: Saturation adjusted image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be PIL Image. Got {}".format(type(img)))

    im = img.astype(np.float32)
    degenerate = cv2.cvtColor(cv2.cvtColor(im, cv2.COLOR_RGB2GRAY), cv2.COLOR_GRAY2RGB)
    im = (1 - saturation_factor) * degenerate + saturation_factor * im
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_hue(img, hue_factor):
    """Adjust hue of an image.
    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.
    `hue_factor` is the amount of shift in H channel and must be in the
    interval `[-0.5, 0.5]`.
    See https://en.wikipedia.org/wiki/Hue for more details on Hue.
    Args:
        img (np.ndarray): CV Image to be adjusted.
        hue_factor (float):  How much to shift the hue channel. Should be in
            [-0.5, 0.5]. 0.5 and -0.5 give complete reversal of hue channel in
            HSV space in positive and negative direction respectively.
            0 means no shift. Therefore, both -0.5 and 0.5 will give an image
            with complementary colors while 0 gives the original image.
    Returns:
        np.ndarray: Hue adjusted image.
    """
    if not (-0.5 <= hue_factor <= 0.5):
        raise ValueError("hue_factor is not in [-0.5, 0.5].")

    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    im = img.astype(np.uint8)
    hsv = cv2.cvtColor(im, cv2.COLOR_RGB2HSV_FULL)
    hsv[..., 0] += np.uint8(hue_factor * 255)

    im = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB_FULL)
    return im.astype(img.dtype)


def adjust_gamma(img, gamma, gain=1):
    """Perform gamma correction on an image.
    Also known as Power Law Transform. Intensities in RGB mode are adjusted
    based on the following equation:
        I_out = 255 * gain * ((I_in / 255) ** gamma)
    See https://en.wikipedia.org/wiki/Gamma_correction for more details.
    Args:
        img (np.ndarray): CV Image to be adjusted.
        gamma (float): Non negative real number. gamma larger than 1 make the
            shadows darker, while gamma smaller than 1 make dark regions
            lighter.
        gain (float): The constant multiplier.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    if gamma < 0:
        raise ValueError("Gamma should be a non-negative real number")

    im = img.astype(np.float32)
    im = 255.0 * gain * np.power(im / 255.0, gamma)
    im = im.clip(min=0.0, max=255.0)
    return im.astype(img.dtype)


def to_grayscale(img, num_output_channels=1):
    """Convert image to grayscale version of image.
    Args:
        img (np.ndarray): Image to be converted to grayscale.
    Returns:
        CV Image:  Grayscale version of the image.
                    if num_output_channels == 1 : returned image is single channel
                    if num_output_channels == 3 : returned image is 3 channel with r == g == b
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))

    if num_output_channels == 1:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    elif num_output_channels == 3:
        img = cv2.cvtColor(cv2.cvtColor(img, cv2.COLOR_RGB2GRAY), cv2.COLOR_GRAY2RGB)
    else:
        raise ValueError("num_output_channels should be either 1 or 3")

    return img


def gaussian_noise(img: np.ndarray, mean, std):
    imgtype = img.dtype
    gauss = np.random.normal(mean, std, img.shape).astype(np.float32)
    noisy = np.clip((1 + gauss) * img.astype(np.float32), 0, 255)
    return noisy.astype(imgtype)


def poisson_noise(img):
    imgtype = img.dtype
    img = img.astype(np.float32) / 255.0
    vals = len(np.unique(img))
    vals = 2 ** np.ceil(np.log2(vals))
    noisy = 255 * np.clip(
        np.random.poisson(img.astype(np.float32) * vals) / float(vals), 0, 1
    )
    return noisy.astype(imgtype)


def salt_and_pepper(img, prob=0.01):
    """Adds "Salt & Pepper" noise to an image.
    prob: probability (threshold) that controls level of noise
    """
    imgtype = img.dtype
    rnd = np.random.rand(img.shape[0], img.shape[1])
    noisy = img.copy()
    noisy[rnd < prob / 2] = 0.0
    noisy[rnd > 1 - prob / 2] = 255.0
    return noisy.astype(imgtype)


def cv_transform(img):
    img = salt_and_pepper(img)
    return to_tensor(img)


def pil_transform(img):
    return functional.to_tensor(img)



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from PIL import Image
from torchvision.models.resnet import BasicBlock, conv1x1
from torchvision.ops.boxes import box_area


def remove_padding(seq):
    r"""
    Remove the trailing zeros from the provided input

    Parameters
    ----------
    list: List of integers
        Predicted sequence

    Returns
    -------
    list: List of integers
        The part of the input before the zero padding

    """
    pad_len = 0
    for x in reversed(seq):
        if x != 0:
            break
        pad_len += 1
    if pad_len == 0:
        return seq, 0

    un_padded = seq[:-pad_len]
    return un_padded, pad_len


def probabilities_to_predictions(probabilities):
    r"""
    Convert probabilities to predictions

    Parameters
    ----------
    probabilities : Tensor[batch_size, vocab_size, seq_len]
        All log probabilities coming out at the last stage of the decoder

    Returns
    -------
    predictions : tensor [batch_size, output_sequence_length]
        The prediceted trags

    """
    # max_idx: [batch_size, seq_len]
    max_idx = torch.argmax(probabilities, dim=1)
    return max_idx


def print_target_predict(target, predictions, filenames=None, batch_idx=0):
    r"""
    For the Tags, print the target and predicted tensors for the specified batch index

    We expect to have the batch size as the first dimension.
    Only the specified batch is extractred and the remaining dimenions are flattened.
    The results are printed as 2 lists with the target on top and the predictions below underlined

    Parameters
    ---------
    target : tensor [batch_size, output_sequence_length]
        The ground truth tags

    predictions : tensor [batch_size, output_sequence_length]
        The prediceted trags

    filenames : list of string
        The actual filename that provides the data

    batch_idx : int
        Which index in the batch dimension will be printed
    """
    target_flat = target[batch_idx].flatten()
    predictions_flat = predictions[batch_idx].flatten()
    target_label = "target"
    predict_label = "predict"
    if filenames is not None:
        target_label = filenames[batch_idx]
    label_len = max(len(target_label), len(predict_label))
    print("{}: {}".format(target_label.ljust(label_len, " "), target_flat.tolist()))
    print(
        "{}: {}".format(predict_label.ljust(label_len, " "), predictions_flat.tolist())
    )


def load_image(full_fn):
    r"""
    Load an image from the disk as a numpy array

    Parameters
    ----------
    full_fn : string
        The full path filename of the image

    Results
    -------
    img : numpy array: (channels, width, height)
        The loaded image as a numpy array
    """
    with Image.open(full_fn) as f:
        img = np.asarray(f)  # (width, height, channels)
        img = img.transpose(2, 0, 1)  # (channels, width, height)
    return img


def resnet_block(stride=1):
    layers = []
    downsample = nn.Sequential(
        conv1x1(256, 512, stride),
        nn.BatchNorm2d(512),
    )
    layers.append(BasicBlock(256, 512, stride, downsample))
    layers.append(BasicBlock(512, 512, 1))
    return nn.Sequential(*layers)


def repackage_hidden(h):
    r"""
    Wraps hidden states in new Tensors, to detach them from their history.
    """
    if isinstance(h, torch.Tensor):
        return h.detach()
    else:
        return tuple(repackage_hidden(v) for v in h)


def accuracy(scores, targets, k):
    """
    Computes top-k accuracy, from predicted and true labels.

    :param scores: scores from the model
    :param targets: true labels
    :param k: k in top-k accuracy
    :return: top-k accuracy
    """

    batch_size = targets.size(0)
    _, ind = scores.topk(k, 1, True, True)
    correct = ind.eq(targets.view(-1, 1).expand_as(ind))
    correct_total = correct.view(-1).float().sum()  # 0D tensor
    return correct_total.item() * (100.0 / batch_size)


def clip_gradient(optimizer, grad_clip):
    """
    Clips gradients computed during backpropagation to avoid explosion of gradients.

    :param optimizer: optimizer with the gradients to be clipped
    :param grad_clip: clip value
    """
    for group in optimizer.param_groups:
        for param in group["params"]:
            if param.grad is not None:
                param.grad.data.clamp_(-grad_clip, grad_clip)


class AverageMeter(object):
    """
    Keeps track of most recent, average, sum, and count of a metric.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


@torch.no_grad()
def bip_accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    if target.numel() == 0:
        return [torch.zeros([], device=output.device)]
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h), (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)


def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2, (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)


# modified from torchvision to also return the union
def box_iou(boxes1, boxes2):
    area1 = box_area(boxes1)
    area2 = box_area(boxes2)

    lt = torch.max(boxes1[:, None, :2], boxes2[:, :2])  # [N,M,2]
    rb = torch.min(boxes1[:, None, 2:], boxes2[:, 2:])  # [N,M,2]

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    inter = wh[:, :, 0] * wh[:, :, 1]  # [N,M]

    union = area1[:, None] + area2 - inter

    iou = inter / union
    return iou, union


def generalized_box_iou(boxes1, boxes2):
    """
    Generalized IoU from https://giou.stanford.edu/

    The boxes should be in [x0, y0, x1, y1] format

    Returns a [N, M] pairwise matrix, where N = len(boxes1)
    and M = len(boxes2)
    """
    # degenerate boxes gives inf / nan results
    # so do an early check
    assert (boxes1[:, 2:] >= boxes1[:, :2]).all()
    assert (boxes2[:, 2:] >= boxes2[:, :2]).all()
    iou, union = box_iou(boxes1, boxes2)

    lt = torch.min(boxes1[:, None, :2], boxes2[:, :2])
    rb = torch.max(boxes1[:, None, 2:], boxes2[:, 2:])

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    area = wh[:, :, 0] * wh[:, :, 1]

    return iou - (area - union) / area


class MLP(nn.Module):
    """Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(
            nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim])
        )

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x


def generate_square_subsequent_mask(sz: int, device: str = "cpu") -> torch.Tensor:
    """Generate the attention mask for causal decoding"""
    mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
    mask = (
        mask.float()
        .masked_fill(mask == 0, float("-inf"))
        .masked_fill(mask == 1, float(0.0))
    ).to(device=device)
    return mask


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience.
    Source from: https://github.com/Bjarten/early-stopping-pytorch
    """

    def __init__(self, patience=2, verbose=False, delta=0, trace_func=print):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
            trace_func (function): trace print function.
                            Default: print
        """
        self._patience = patience
        self._verbose = verbose
        self._counter = 0
        self._best_score = None
        self._early_stop = False
        self._val_loss_min = np.Inf
        self._delta = delta
        self._trace_func = trace_func

    def __call__(self, val_loss):
        score = -val_loss
        save_checkpoint = True
        if self._best_score is None:
            self._best_score = score
            save_checkpoint = True
            if self._verbose:
                verb = f"Validation loss decreased ({self._val_loss_min:.6f} --> {val_loss:.6f})."
                self._trace_func(verb)
            self._val_loss_min = val_loss
        elif score < self._best_score + self._delta:
            self._counter += 1
            self._trace_func(
                f"EarlyStopping counter: {self._counter} out of {self._patience}"
            )
            if self._counter >= self._patience:
                self._early_stop = True
                save_checkpoint = False
        else:
            self._best_score = score
            save_checkpoint = True
            self._counter = 0
            if self._verbose:
                verb = f"Validation loss decreased ({self._val_loss_min:.6f} --> {val_loss:.6f})."
                self._trace_func(verb)
            self._val_loss_min = val_loss
        return save_checkpoint


def print_dict(m: dict):
    r"""
    Print dict elements in separate lines sorted by keys
    """
    if len(m) == 0:
        return

    # Check if the key is a stringified integer
    first_key = next(iter(m))
    is_numeric = isinstance(first_key, str) and first_key.isnumeric()
    if is_numeric:
        keys = sorted([int(k) for k in m.keys()])
    else:
        keys = sorted([k for k in m.keys()])

    for k in keys:
        if is_numeric:
            v = m[str(k)]
        else:
            v = m[k]
        print("{}: {}".format(k, v))


def print_list(lst: list):
    r"""
    Print list elements in separate lines
    """
    for i, elm in enumerate(lst):
        if isinstance(elm, list):
            print("{}: ({}) - {}".format(i, len(elm), elm))
        else:
            print("{}: {}".format(i, elm))



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import glob
import os

import numpy as np
from PIL import Image

import docling_ibm_models.tableformer.common as c
from docling_ibm_models.tableformer.data_management.data_transformer import (
    DataTransformer,
)


def dump_np(img_np: np.array, fn, n=6):
    # Expect to receive a numpy array for an image with the shape [channels, rows, columns]
    s = img_np.shape
    if s[0] not in [1, 2, 3, 4] or len(s) != 3:
        print("Image of invalid shape: {}".format(s))
        return

    channels = s[0]
    rows = s[1]
    cols = s[2]
    w = n + 6
    with open(fn, "w") as fd:
        for r in range(rows):
            for col in range(cols):
                for ch in range(channels):
                    x = img_np[ch][r][col]
                    if isinstance(x, np.float32):
                        f_str = "0:>{}.{}f".format(w, n)
                    elif isinstance(x, np.uint8):
                        f_str = "0:>{}".format(w)
                    else:
                        return False

                    x_str = ("{" + f_str + "}").format(x)
                    fd.write(x_str)
                    if ch < channels - 1:
                        fd.write(" ")
                fd.write("\n")
    return True


def dump_channels(save_dir, fn_prefix, img_np: np.array):
    # Dump the np array into 3 files per channel
    img_np_ch0 = img_np[0, :, :]
    img_np_ch1 = img_np[1, :, :]
    img_np_ch2 = img_np[2, :, :]
    txt_ch0_fn = os.path.join(save_dir, fn_prefix + "_ch0.txt")
    txt_ch1_fn = os.path.join(save_dir, fn_prefix + "_ch1.txt")
    txt_ch2_fn = os.path.join(save_dir, fn_prefix + "_ch2.txt")
    np.savetxt(txt_ch0_fn, img_np_ch0)
    np.savetxt(txt_ch1_fn, img_np_ch1)
    np.savetxt(txt_ch2_fn, img_np_ch2)
    print(f"{txt_ch0_fn}")
    print(f"{txt_ch1_fn}")
    print(f"{txt_ch2_fn}")


def prepare_image(config):
    transformer = DataTransformer(config)
    predict_dir = config["predict"]["predict_dir"]
    use_normalization = config["dataset"]["image_normalization"]["state"]

    pattern = os.path.join(predict_dir, "*.png")
    for img_fn in glob.glob(pattern):
        print(f"img_fn: {img_fn}")

        with Image.open(img_fn) as img:
            # Dump the initial image in txt files
            img_np = np.array(img)

            # Reshape the image in order to print it
            img_np_m = np.moveaxis(img_np, 2, 0)
            print(
                "orig. img_np.shape: {}, reshaped image: {}".format(
                    img_np.shape, img_np_m.shape
                )
            )
            original_fn = img_fn + "_python.txt"
            dump_np(img_np_m, original_fn)

            r_img_ten = transformer.rescale_in_memory(img, use_normalization)
            print("npimgc: {} - {}".format(r_img_ten.type(), r_img_ten.size()))

            # Dump the processed image tensor in txt files
            r_img_np = r_img_ten.numpy()

            prepared_fn = img_fn + "_python_prepared.txt"
            dump_np(r_img_np, prepared_fn)


if __name__ == "__main__":
    config = c.parse_arguments()
    prepare_image(config)



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import argparse
import json
import logging
import os

import torch

import docling_ibm_models.tableformer.settings as s
from docling_ibm_models.tableformer.models.common.base_model import BaseModel

LOG_LEVEL = logging.DEBUG
logger = s.get_custom_logger("common", LOG_LEVEL)


def validate_config(config):
    r"""
    Validate the provided configuration file.
    A ValueError exception will be thrown in case the config file is invalid

    Parameters
    ----------
    config : dictionary
        Configuration for the tablemodel

    Returns
    -------
    bool : True on success
    """
    if "model" not in config:
        return True
    if "preparation" not in config:
        return True
    assert (
        "max_tag_len" in config["preparation"]
    ), "Config error: 'preparation.max_tag_len' parameter is missing"
    if "seq_len" in config["model"]:
        assert (
            config["model"]["seq_len"] > 0
        ), "Config error: 'model.seq_len' should be positive"
        assert config["model"]["seq_len"] <= (
            config["preparation"]["max_tag_len"] + 2
        ), "Config error: 'model.seq_len' should be up to 'preparation.max_tag_len' + 2"

    return True


def parse_arguments():
    r"""
    Parse the input arguments
    A ValueError exception will be thrown in case the config file is invalid
    """
    parser = argparse.ArgumentParser(description="Train the TableModel")
    parser.add_argument(
        "-c", "--config", required=True, default=None, help="configuration file (JSON)"
    )
    args = parser.parse_args()
    config_filename = args.config

    assert os.path.isfile(config_filename), "FAILURE: Config file not found."
    return read_config(config_filename)


def read_config(config_filename):
    with open(config_filename, "r") as fd:
        config = json.load(fd)

    # Validate the config file
    validate_config(config)

    return config


def safe_get_parameter(input_dict, index_path, default=None, required=False):
    r"""
    Safe get parameter from a nested dictionary.

    Provide a nested dictionary (dictionary of dictionaries) and a list of indices:
    - If the whole index path exists the value pointed by it is returned
    - Otherwise the default value is returned.

    Input:
        input_dict: Data structure of nested dictionaries.
        index_path: List with the indices path to follow inside the input_dict.
        default: Default value to return if the indices path is broken.
        required: If true a ValueError exception will be raised in case the parameter does not exist
    Output:
        The value pointed by the index path or "default".
    """
    if input_dict is None or index_path is None:
        return default

    d = input_dict
    for i in index_path[:-1]:
        if i not in d:
            if required:
                raise ValueError("Missing parameter: {}".format(i))
            return default
        d = d[i]

    last_index = index_path[-1]
    if last_index not in d:
        if required:
            raise ValueError("Missing parameter: {}".format(last_index))
        return default

    return d[last_index]


def get_prepared_data_filename(prepared_data_part, dataset_name):
    r"""
    Build the full filename of the prepared data part

    Parameters
    ----------
    prepared_data_part : string
        Part of the prepared data
    dataset_name : string
        Name of the dataset

    Returns
    -------
    string
        The full filename for the prepared file
    """
    template = s.PREPARED_DATA_PARTS[prepared_data_part]
    if "<POSTFIX>" in template:
        template = template.replace("<POSTFIX>", dataset_name)
    return template


def create_dataset_and_model(config, purpose, fixed_padding=False):
    r"""
    Gets a model from configuration

    Parameters
    ---------
    config : Dictionary
        The configuration of the model
    purpose : string
        One of "train", "eval", "predict"
    fixed_padding : bool
        Parameter passed to the constructor of the DataLoader

    Returns
    -------
    In case a Model cannot be initialized return None, None, None. Otherwise:

    device : selected device
    dataset : Instance of the DataLoader
    model : Instance of the model
    """
    from docling_ibm_models.tableformer.data_management.tf_dataset import TFDataset

    model_type = config["model"]["type"]
    model = None

    # Get env vars:
    use_cpu_only = os.environ.get("USE_CPU_ONLY", False)
    use_cuda_only = not use_cpu_only

    # Use the cpu for the evaluation
    device = "cpu"  # Default, run on CPU
    num_gpus = torch.cuda.device_count()  # Check if GPU is available
    if use_cuda_only:
        device = "cuda:0" if num_gpus > 0 else "cpu"  # Run on first available GPU
    else:
        device = "cpu"

    # Create the DataLoader
    # loader = DataLoader(config, purpose, fixed_padding=fixed_padding)
    dataset = TFDataset(config, purpose, fixed_padding=fixed_padding)
    dataset.set_device(device)
    dataset_val = None
    if config["train"]["validation"] and purpose == "train":
        dataset_val = TFDataset(config, "val", fixed_padding=fixed_padding)
        dataset_val.set_device(device)
    if model_type == "TableModel04_rs":
        from docling_ibm_models.tableformer.models.table04_rs.tablemodel04_rs import (  # noqa: F401
            TableModel04_rs,
        )
    # Find the model class and create an instance of it
    for candidate in BaseModel.__subclasses__():
        if candidate.__name__ == model_type:
            init_data = dataset.get_init_data()
            model = candidate(config, init_data, purpose, device)

    if model is None:
        logger.warn("Not found model: " + str(model_type))
        return None, None, None

    logger.info("Found model: " + str(model_type))

    if purpose == s.PREDICT_PURPOSE:
        return device, dataset, model
    else:
        return device, dataset, dataset_val, model



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging

import numpy as np

import docling_ibm_models.tableformer.settings as s

LOG_LEVEL = logging.INFO


class MyWelford:
    r"""
    Running computation of the sample mean and sample variance using Welford's algorithm
    """

    def __init__(self):
        self._i = 0  # Running index
        self._m = 0  # Running mean
        self._s = 0  # (n - 1) * variance

    def reset(self):
        r"""
        Reset the object
        """
        self._i = 0
        self._m = 0
        self._s = 0

    def add(self, xi):
        r"""
        Invoke add each time a new sample arrives

        Inputs:
           xi: The next sample of data
        """
        self._i += 1
        old_m = self._m
        self._m = self._m + (xi - self._m) / self._i
        self._s = self._s + (xi - self._m) * (xi - old_m)

    def results(self):
        r"""
        Get the computed mean, variance and standard deviation up to now

        Outputs:
            m: Sample mean
            v: Sample variance
            std: Sample standard deviation
        """
        if self._i <= 1:
            return None, None, None

        # v = self._s / (self._i - 1)  # Sample variance
        v = self._s / (self._i)  # Population variance
        std = np.sqrt(v)
        return self._m, v, std


class MyWelfordImg(MyWelford):
    r"""
    Welford algorithm to calculate running mean and sample variance for images
    """

    def __init__(self):
        super(MyWelfordImg, self).__init__()

    def add(self, img):
        r"""
        Input:
            img: An image numpy array (channel, width, height). The only requirement is to have the
                 channels as the first dimension and have 3 dimensions in total
        """
        channels = img.shape[0]
        flat_dim = img.shape[1] * img.shape[2]
        img_r = img.reshape(channels, flat_dim)

        for i in range(flat_dim):
            super(MyWelfordImg, self).add(img_r[:, i])


class ChanVarianceImg:
    r"""
    Chan's algorithm to compute a running variance with support of sub-samples
    In this implementation each sub-sample is an images

    Math for the original paper:
    https://github.ibm.com/nli/variance_formulae
    """

    def __init__(self):
        r""" """
        self._first = True
        # Size of the calculated dataset
        self._n = 0
        # Sum of the samples for the 3 image channels
        self._t = 0
        # Sum of the square differences of the deviations of the samples from the mean
        self._s = 0

    def add(self, img):
        r"""
        Add the provided image to the computation of the dataset statistics

        Input:
            img: An image numpy array (channel, width, height). The only requirement is to have the
                 channels as the first dimension and have 3 dimensions in total
        """
        ch = img.shape[0]
        n = img.shape[1] * img.shape[2]
        img = img.reshape(ch, n)
        img_t = img.sum(axis=1)
        img_t_v = img_t.reshape(ch, 1)
        diff = (img - (img_t_v / n)) ** 2
        img_s = diff.sum(axis=1)

        if not self._first:
            c = (self._n / (n * (self._n + n))) * (
                ((n / self._n) * self._t - img_t) ** 2
            )
            self._s += img_s + c
            self._t += img_t
        else:
            self._s = img_s
            self._t = img_t
            self._first = False
        self._n += n

    def results(self):
        r"""
        Get the computed statistics

        Output:
            mean: Mean for the complete dataset
            var: Population variance for the complete dataset
            std: Population standard deviation for the complete dataset
        """
        mean = list(self._t / self._n)
        var = list(self._s / self._n)  # Population variance
        std = list(np.sqrt(var))

        return mean, var, std

    def reset(self):
        r"""
        Reset the object to start over again
        """
        self._n = 0
        self._t = 0
        self._s = 0
        self._first = True


if __name__ == "__main__":
    logger = s.get_custom_logger("variance", LOG_LEVEL)

    n = 50000
    channels = 3
    width = 448
    height = 448

    my = ChanVarianceImg()
    # Generate random images
    for i in range(n):
        logger.info(i)
        img = 255 * np.random.rand(channels, width, height)
        my.add(img)

    # Calculate the statistics
    m, v, std = my.results()
    assert m.shape == (3,), "Wrong mean dimension"
    assert v.shape == (3,), "Wrong variance dimension"
    assert std.shape == (3,), "Wrong std dimension"



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import glob
import logging
import os
import time
from abc import ABC, abstractmethod
from pathlib import Path

import torch

import docling_ibm_models.tableformer.settings as s

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


class BaseModel(ABC):
    r"""
    BaseModel provides some common functionality for all models:
    - Saves checkpoint files for each epoch
    - Loads the model from the best available checkpoint
    - Save repository branch and commit
    """

    def __init__(self, config, init_data, device):
        r"""
        Inputs:
            config: The configuration file
            init_data: Dictionary with initialization data. This dictionary can be used to pass any
                       kind of initialization data for the models
            device: The device used to move the tensors of the model
        """
        super(BaseModel, self).__init__()

        # Set config and device
        self._config = config
        self._init_data = init_data

        self._device = device

        self._save_dir = config["model"]["save_dir"]
        self._load_checkpoint = None
        if "load_checkpoint" in config["model"]:
            self._load_checkpoint = config["model"]["load_checkpoint"]

        self._branch_name = "dev/next"
        self._commit_sha = "1"

        # Keep a dictionary with the starting times per epoch.
        # NOTICE: Epochs start from 0
        self._epoch_start_ts = {0: time.time()}

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    @abstractmethod
    def predict(self, img, max_steps, beam_size, return_attention):
        pass

    def count_parameters(self):
        r"""Counts the number of trainable parameters of this model

        Output:
            num_parameters: number of trainable parameters
        """
        num_parameters = sum(p.numel() for p in self.parameters() if p.requires_grad)

        return num_parameters

    def get_code_version(self):
        r"""Gets the source control version of this model code

        Returns
        -------
        branch_name : str
            The name of the Git branch of this model code
        commit_sha : str
            The unique identifier of the Git commit of this model code
        """

        return self._branch_name, self._commit_sha

    def get_save_directory(self):
        r"""
        Return the save directory
        """
        return self._save_dir

    def is_saved(self):
        r"""
        This method returns True if both conditions are met:
        1. There is a checkpoint file for the model.
        2. The checkpoint file corresponds to the last training epoch set in the configuration file.
        """
        # Get the saved_model
        saved_model, _ = self._load_best_checkpoint()

        if saved_model is None:
            return False

        epochs = self._config["train"]["epochs"]
        self._log().debug(
            "Best epoch in saved model: {}; Number of epochs in config: {}".format(
                saved_model["epoch"], epochs
            )
        )
        if epochs == saved_model["epoch"] + 1:
            return True

        return False

    def save(self, epoch=None, optimizers=None, losses=None, model_parameters=None):
        r"""
        Save the model data to the disk as a pickle file.

        Parameters
        ----------
        epoch: Training epoch
        optimizers: Dictionary with the optimizers. The key specifies what the optimizer is
                     used for. The 'state_dict' of each optimizer will be saved in the
                     checkpoint file.
        losses: Dictionary with the losses. The key specifies what the loss is used for. Each
                value is a list
        model_parameters: Dictionary with model specific parameters that we need to save in the
                              checkpoint file.
        Returns
        -------
        True if success, False otherwise
        """
        # Get the checkpoint_filename
        c_filename = self._build_checkpoint_filename(epoch)
        self._log().debug("Trying to save checkpoint file: {}".format(c_filename))

        # Prepare a dictionary with all data we want to save
        optimizers_state_dict = None
        if optimizers is not None:
            optimizers_state_dict = {k: v.state_dict() for k, v in optimizers.items()}

        model_data = {
            "model_state_dict": self.state_dict(),
            "epoch": epoch,
            "optimizers": optimizers_state_dict,
            "losses": losses,
            "model_parameters": model_parameters,
        }

        # Add the processing time per epoch
        now = time.time()
        self._epoch_start_ts[epoch + 1] = now
        if epoch in self._epoch_start_ts:
            dt = now - self._epoch_start_ts[epoch]
            model_data["epoch_start_ts"] = self._epoch_start_ts[epoch]
            model_data["epoch_dt"] = dt

        # Create the save directory
        Path(self._save_dir).mkdir(parents=True, exist_ok=True)

        # Save the model
        torch.save(model_data, c_filename)

        # Return true if file is present, otherwise false
        if not os.path.isfile(c_filename):
            self._log().error("Cannot find the file to save: " + c_filename)
            return False

        # store code branch name and commit
        version_file = os.path.join(self._save_dir, "_version")
        with open(version_file, "w") as text_file:
            print("Model is using code [commit:branch]", file=text_file)
            print("{}:{}".format(self._commit_sha, self._branch_name), file=text_file)

        return True

    def load(self, optimizers=None):
        r"""
        Load the model data from the disk.
        The method will iterate over all *.check files and try to load the one from the highest
        epoch.

        Input:
            -optimizers: Dictionary with optimizers. If it is not null the keys will be used to
                         associate the corresponding state_dicts from the checkpoint file and update
                         the internal states of the provided optimizers.

        Output:
            - Success: True/ False
            - epoch: Loaded epoch or -1 if there are no checkpoint files
            - optimizers: Dictionary with loaded optimizers or empty dictionary of there is no
                          checkpoint file
            - losses: Dictionary with loaded losses or empty dictionary of there is no checkpoint
                      file
            - model_parameters: Dictionary with the model parameters or empty dictionary if there
                                are no checkpoint files
        """
        # Get the saved_model
        saved_model, _ = self._load_best_checkpoint()

        # Restore the model
        if saved_model is None:
            self._log().debug("No saved model checkpoint found")
            return False, -1, optimizers, {}, {}

        self._log().debug("Loading model from checkpoint file")
        self.load_state_dict(saved_model["model_state_dict"])

        epoch = 0
        if "epoch" in saved_model:
            epoch = saved_model["epoch"]
        losses = {}
        if "losses" in saved_model:
            losses = saved_model["losses"]
        model_parameters = saved_model["model_parameters"]

        if optimizers is not None:
            for key, optimizer_state_dict in saved_model["optimizers"].items():
                optimizers[key].load_state_dict(optimizer_state_dict)

        # Reset the start_ts of the next epoch
        self._epoch_start_ts[epoch + 1] = time.time()

        return True, epoch, optimizers, losses, model_parameters

    def _load_best_checkpoint(self):
        r"""
        If a "load_checkpoint" file has been provided, load this one.
        Otherwise use the "save_dir" and load the one with the most advanced epoch

        Returns
        -------
        saved_model : dictionary
            Checkpoint file contents generated by torch.load, or None
        checkpoint_file : string
            Filename of the loaded checkpoint, or None
        """
        checkpoint_files = []
        # If a "load_checkpoint" file is provided, try to load it
        if self._load_checkpoint is not None:
            if not os.path.exists(self._load_checkpoint):
                self._log().error(
                    "Cannot load the checkpoint: {}".format(self._load_checkpoint)
                )
                return None, None
            checkpoint_files.append(self._load_checkpoint)
        else:
            # Iterate over all check files from the directory by reverse alphabetical order
            # This will get the biggest epoch first
            checkpoint_files = glob.glob(os.path.join(self._save_dir, "*.check"))
            checkpoint_files.sort(reverse=True)

        for checkpoint_file in checkpoint_files:
            try:
                # Try to load the file
                self._log().info(
                    "Loading model checkpoint file: {}".format(checkpoint_file)
                )
                saved_model = torch.load(checkpoint_file, map_location=self._device)
                return saved_model, checkpoint_file
            except RuntimeError:
                self._log().error("Cannot load file: {}".format(checkpoint_file))

        return None, None

    def _build_checkpoint_filename(self, epoch):
        r"""
        Construct the full path for the filename of this checkpoint
        """
        dataset_name = self._config["dataset"]["name"]
        model_type = self._config["model"]["type"]
        model_name = self._config["model"]["name"]
        filename = "{}_{}_{}_{:03}.check".format(
            model_type, model_name, dataset_name, epoch
        )
        c_filename = os.path.join(self._save_dir, filename)

        return c_filename



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import copy
import logging
import os
import random

import cv2
import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image, ImageDraw

import docling_ibm_models.tableformer.data_management.transforms as T
import docling_ibm_models.tableformer.settings as s

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


class DataTransformer:
    r"""
    Data transformations for the images and bboxes

    Check the "help" fields inside the config file for an explanation of each parameter
    """

    def __init__(self, config):
        self._config = config

        print("DataTransformer Init!")

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def append_id(self, filename):
        name, ext = os.path.splitext(filename)
        return "{name}_{uid}{ext}".format(name=name, uid="resized", ext=ext)

    def load_image(self, img_fn):
        r"""
        Load an image from the disk

        Parameters
        ----------
            img_fn: The filename of the image

        Returns
        -------
            PIL image object
        """

        img = Image.open(img_fn)
        return img

    def load_image_cv2(self, img_fn):
        r"""Load an image from the disk

        Parameters
        ----------
            img_fn: The filename of the image

        Returns
        -------
            CV2  image object
        """
        img = cv2.imread(img_fn)
        return img

    def save_image(self, img, img_fn):
        img.save(self.append_id(img_fn))

    def renderbboxes(self, img, bboxes):
        draw_img = ImageDraw.Draw(img)
        for i in range(len(bboxes)):
            draw_img.rectangle(bboxes[i], fill=None, outline=(255, 0, 0))
        return img

    def get_dataset_settings(self):
        dataset = {}
        debug = {"save_debug_images": False}

        if "dataset" in self._config:
            dataset = self._config["dataset"]
        if "debug" in self._config:
            debug = self._config["debug"]

        return dataset, debug

    def _prepare_image_from_file(self, image_fn, bboxes, convert_box=True):
        r"""
        Load the image from file and prepare it

        Parameters
        ----------
        image_fn : string
            Filename to load the image
        bboxes : dict
            Bounding boxes of the image
        convert_box : bool
            If true the bboxes are converted to xcycwh format

        Returns
        -------
        PIL image
            A PIL image object with the image prepared according to the settings in the config file
        bboxes : dict
            Converted bboxes of the image
        """
        im = self.load_image(image_fn)
        return self._prepare_image(im, bboxes, convert_box, image_fn)

    def _prepare_image(self, im, bboxes, convert_box=True, image_fn=None):
        r"""
        Parameters
        ----------
        im : PIL image object
        bboxes : dict
            Bounding boxes of the image
        convert_box : bool
            If true the bboxes are converted to xcycwh format
        image_fn : string
            Filename of the original image or None. It is used to save augmented image for debugging

        Returns
        -------
        PIL image
            A PIL image object with the image prepared according to the settings in the config file
        bboxes : dict
            Converted bboxes of the image
        """
        debug_settings = False
        settings, debug_settings = self.get_dataset_settings()

        desired_size = settings["resized_image"]
        old_size = im.size

        # Calculate Aspect Ratio if needed
        if settings["keep_AR"]:
            ratio = float(desired_size) / max(old_size)
        else:
            ratio = 1  # Square image

        new_size = old_size
        if max(old_size) < desired_size:
            # Image is smaller than desired
            # Upscale?
            if settings["up_scaling_enabled"]:
                # Calculate new image size, taking into account aspect ratio
                new_size = tuple([int(x * ratio) for x in old_size])
            else:
                new_size = old_size
        else:
            if settings["keep_AR"]:
                new_size = tuple([int(x * ratio) for x in old_size])
            else:
                new_size = [desired_size, desired_size]

        if not settings["keep_AR"]:
            if settings["up_scaling_enabled"]:
                new_size = [desired_size, desired_size]

        ######################################################################################
        # Use OpenCV to resize the image
        #
        # im = im.resize(new_size, Image.ANTIALIAS)

        import cv2

        np_im = np.array(im)
        np_resized = cv2.resize(np_im, new_size, interpolation=cv2.INTER_LANCZOS4)
        im = Image.fromarray(np_resized)
        ######################################################################################

        new_bboxes = copy.deepcopy(bboxes)

        # Resize bboxes (in pixels)
        x_scale = new_size[0] / old_size[0]
        y_scale = new_size[1] / old_size[1]
        # loop over bboxes
        for i in range(len(new_bboxes)):
            new_bboxes[i][0] = x_scale * bboxes[i][0]
            new_bboxes[i][1] = y_scale * bboxes[i][1]
            new_bboxes[i][2] = x_scale * bboxes[i][2]
            new_bboxes[i][3] = y_scale * bboxes[i][3]

        # Set background color for padding
        br = settings["padding_color"][0]
        bg = settings["padding_color"][1]
        bb = settings["padding_color"][2]
        bcolor = (br, bg, bb)
        # Create empty canvas of background color and desired size
        new_im = Image.new(mode="RGB", size=(desired_size, desired_size), color=bcolor)

        if "grayscale" in settings:
            if settings["grayscale"]:
                im = im.convert("LA")

        if settings["padding_mode"] == "frame":
            # If paddinds are around image, paste resized image in the center
            x_pad = (desired_size - new_size[0]) // 2
            y_pad = (desired_size - new_size[1]) // 2
            # Paste rescaled image
            new_im.paste(im, (x_pad, y_pad))
            # Offset (pad) bboxes
            # loop over bboxes
            for i in range(len(new_bboxes)):
                new_bboxes[i][0] += x_pad
                new_bboxes[i][1] += y_pad
                new_bboxes[i][2] += x_pad
                new_bboxes[i][3] += y_pad
        else:
            # Otherwise paste in the 0,0 coordinates
            new_im.paste(im, (0, 0))

        if debug_settings:
            if debug_settings["save_debug_images"]:
                aug_im = self.renderbboxes(new_im, bboxes)
                if "grayscale" in settings:
                    if settings["grayscale"]:
                        aug_im = aug_im.convert("LA")
                self.save_image(aug_im, image_fn)
        if convert_box:
            bboxes = self.xyxy_to_xcycwh(new_bboxes, desired_size)
        return new_im, bboxes

    # convert bboxes from [x1, y1, x2, y2] format to [xc, yc, w, h] format
    def xyxy_to_xcycwh(self, bboxes, size):
        # Use the "dataset.bbox_format" parameter to decide which bbox format to use
        bbox_format = self._config["dataset"].get("bbox_format", "4plet")

        conv_bboxes = []
        for i in range(len(bboxes)):
            x1 = bboxes[i][0] / size  # X1
            y1 = bboxes[i][1] / size  # Y1
            x2 = bboxes[i][2] / size  # X2
            y2 = bboxes[i][3] / size  # Y2
            xc = (x1 + x2) / 2
            yc = (y1 + y2) / 2
            bw = abs(x2 - x1)
            bh = abs(y2 - y1)

            if bbox_format == "5plet":
                cls = bboxes[i][4]
                conv_bboxes.append([xc, yc, bw, bh, cls])
            else:
                conv_bboxes.append([xc, yc, bw, bh])

        # conv_bboxes = bboxes
        return conv_bboxes

    def rescale_in_memory(self, image, normalization):
        r"""
        Receive image and escale it in memory

        Parameters
        ----------
        image : PIL image
            The image data to rescale
        normalization : dictionary
            The normalization information with the format:
                "state": "true or false if image normalization is to be enabled",
                "mean": "mean values to use if state is true",
                "std": "std values to use if state is true"
        Returns
        -------
        npimgc : FloatTensor
            The loaded and properly transformed image data
        """
        settings, debug_settings = self.get_dataset_settings()
        new_image, _ = self._prepare_image(image, {}, convert_box=False, image_fn=None)
        # Convert to nparray
        npimg = np.asarray(new_image)  # (width, height, channels)

        # Convert to float?
        npimgc = npimg.copy()

        # Transpose numpy array (image)
        npimgc = npimgc.transpose(2, 0, 1)  # (channels, width, height)
        npimgc = torch.FloatTensor(npimgc / 255.0)

        if normalization:
            transform = transforms.Compose(
                [
                    transforms.Normalize(
                        mean=self._config["dataset"]["image_normalization"]["mean"],
                        std=self._config["dataset"]["image_normalization"]["std"],
                    )
                ]
            )
            npimgc = transform(npimgc)

        return npimgc

    def _rescale(self, image_fn, bboxes, normalization):
        r"""
        Rescale, resize, pad the given image and its associated bboxes according to the settings
        from the config

        Parameters
        ----------
        image_fn: full image file name
        bboxes: List with bboxes in the format [x1, y1, x2, y2] with box's top-right,
                bottom-left points
        statistics: Dictionary with statistics over the whole image dataset.
                    The keys are: "mean", "variance", "std" and each value is a list with the
                    coresponding statistical value for each channel. Normally there are 3
                    channels.

        Returns
        -------
        npimgc : FloatTensor
            The loaded and properly transformed image data
        bboxes: List with bboxes in the format [xc, yc, w, h] where xc, yc are the coords of the
                 center, w, h the width and height of the bbox and all are normalized to the
                 scaled size of the image

        Raises
        ------
        ValueError
            In case the configuration and the image dimensions make it impossible to rescale the
            image throw a ValueError exception
        """
        settings, debug_settings = self.get_dataset_settings()
        # new_image is a PIL object
        new_image, new_bboxes = self._prepare_image_from_file(image_fn, bboxes)
        # Convert to nparray
        npimg = np.asarray(new_image)  # (width, height, channels)
        # Convert to float?
        npimgc = npimg.copy()
        # Transpose numpy array (image)
        npimgc = npimgc.transpose(2, 0, 1)  # (channels, width, height)
        npimgc = torch.FloatTensor(npimgc / 255.0)

        if normalization:
            transform = transforms.Compose(
                [
                    transforms.Normalize(
                        mean=self._config["dataset"]["image_normalization"]["mean"],
                        std=self._config["dataset"]["image_normalization"]["std"],
                    )
                ]
            )
            npimgc = transform(npimgc)
        return npimgc, new_bboxes

    def rescale_old(self, image, bboxes, statistics=None):
        r"""
        Rescale, resize, pad the given image and its associated bboxes according to the settings
        from the config

        Input:
            image: np array (channels, width, height)
            bboxes: List with bboxes in the format [x1, y1, x2, y2] with box's top-right,
                    bottom-left points
            statistics: Dictionary with statistics over the whole image dataset.
                        The keys are: "mean", "variance", "std" and each value is a list with the
                        coresponding statistical value for each channel. Normally there are 3
                        channels.

        Output:
            image: np array (channels, resized_image, resized_image)
            bboxes: List with bboxes in the format (xc, yc, w, h) where xc, yc are the coords of the
                     center, w, h the width and height of the bbox and all are normalized to the
                     scaled size of the image

        Exceptions:
            In case the configuration and the image dimensions make it impossible to rescale the
            image throw a ValueError exception
        """
        image_size = 448
        # Convert the image to (width, height, channels)
        image = image.transpose(1, 2, 0)

        # Convert to PIL format and resize
        image = Image.fromarray(image)
        image = image.resize((image_size, image_size), Image.ANTIALIAS)
        return image, bboxes

    def rescale_batch(self, images, bboxes, statistics=None):
        r"""
        Rescale, resize, pad the given batch of images and its associated bboxes according to the
        settings from the config.

        Input:
            images: np array (batch_size, channels, width, height)
            bboxes:
            statistics: Dictionary with statistics over the whole image dataset.
                        The keys are: "mean", "variance", "std" and each value is a list with the
                        coresponding statistical value for each channel. Normally there are 3
                        channels.

        Output:
            image batch: np array (batch_size, channels, resized_image, resized_image)
            bboxes:

        Exceptions:
            In case the configuration and the image dimensions make it impossible to rescale the
            image throw a ValueError exception
        """
        pass

    def sample_preprocessor(self, image_fn, bboxes, purpose, table_bboxes=None):
        r"""
        Rescale, resize, pad the given image and its associated bboxes according to the settings
        from the config

        Parameters
        ----------
        image_fn: full image file name
        bboxes: List with bboxes in the format [x1, y1, x2, y2] with box's top-right,
                bottom-left points
        statistics: Dictionary with statistics over the whole image dataset.
                    The keys are: "mean", "variance", "std" and each value is a list with the
                    coresponding statistical value for each channel. Normally there are 3
                    channels.

        Returns
        -------
        npimgc : FloatTensor
            The loaded and properly transformed image data
        bboxes: List with bboxes in the format [xc, yc, w, h] where xc, yc are the coords of the
                 center, w, h the width and height of the bbox and all are normalized to the
                 scaled size of the image

        Raises
        ------
        ValueError
            In case the configuration and the image dimensions make it impossible to rescale the
            image throw a ValueError exception
        """
        settings, debug_settings = self.get_dataset_settings()
        img = self.load_image_cv2(image_fn)
        img = np.ascontiguousarray(img)

        target = {
            "size": [img.shape[1], img.shape[2]],
            "boxes": (
                torch.from_numpy(np.array(bboxes)[:, :4])
                if purpose != s.PREDICT_PURPOSE
                else None
            ),
            "classes": (
                np.array(bboxes)[:, -1] if purpose != s.PREDICT_PURPOSE else None
            ),
            "area": img.shape[1] * img.shape[2],
        }

        optional_transforms = [T.NoTransformation()]

        # Necessary preprocessing ends here, experimental options begin below.
        # DETR format, might be necessary to keep this structure to share other functions used by
        # the community

        if purpose == s.TRAIN_PURPOSE:
            if self._config["dataset"]["color_jitter"]:
                jitter = T.ColorJitter(
                    brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1
                )
                optional_transforms.append(jitter)

            if self._config["dataset"]["rand_pad"]:
                pad_val = random.randint(0, 50)
                rand_pad = T.RandomPad(pad_val)
                optional_transforms.append(rand_pad)

            if table_bboxes is not None:
                if self._config["dataset"]["rand_crop"]:
                    w_, h_, _ = img.shape[0], img.shape[1], img.shape[2]
                    w_c, h_c = table_bboxes[0], table_bboxes[1]
                    f_w, f_h = random.randint(0, w_c), random.randint(0, h_c)
                    rand_crop = T.RandomCrop((w_, h_), (f_w, f_h))
                    optional_transforms.append(rand_crop)

        # transform_opt = random.choice(optional_transforms)
        normalize = T.Normalize(
            mean=self._config["dataset"]["image_normalization"]["mean"],
            std=self._config["dataset"]["image_normalization"]["std"],
        )
        resized_size = self._config["dataset"]["resized_image"]
        resize = T.Resize([resized_size, resized_size])

        transformations = T.RandomChoice(optional_transforms)

        img, target = transformations(img, target)
        img, target = normalize(img, target)
        img, target = resize(img, target)

        img = img.transpose(2, 1, 0)  # (channels, width, height)
        img = torch.FloatTensor(img / 255.0)
        bboxes_ = target["boxes"]
        classes_ = target["classes"]
        desired_size = img.shape[1]

        if purpose != s.PREDICT_PURPOSE:
            bboxes_ = np.concatenate(
                (bboxes_, np.expand_dims(classes_, axis=1)), axis=1
            )
            bboxes_ = self.xyxy_to_xcycwh(bboxes_, desired_size)

        return img, bboxes_



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging

import torch.nn as nn
import torchvision

import docling_ibm_models.tableformer.settings as s

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


class Encoder04(nn.Module):
    """
    Encoder based on resnet-18
    """

    def __init__(self, enc_image_size, enc_dim=512):
        r"""
        Parameters
        ----------
        enc_image_size : int
                Assuming that the encoded image is a square, this is the length of the image side
        """

        super(Encoder04, self).__init__()
        self.enc_image_size = enc_image_size
        self._encoder_dim = enc_dim

        resnet = torchvision.models.resnet18(pretrained=False)
        modules = list(resnet.children())[:-3]

        self._resnet = nn.Sequential(*modules)
        self._adaptive_pool = nn.AdaptiveAvgPool2d(
            (self.enc_image_size, self.enc_image_size)
        )

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def get_encoder_dim(self):
        return self._encoder_dim

    def forward(self, images):
        """
        Forward propagation
        The encoder_dim 512 is decided by the structure of the image network (modified resnet-19)

        Parameters
        ----------
        images : tensor (batch_size, image_channels, resized_image, resized_image)
                images input

        Returns
        -------
        tensor : (batch_size, enc_image_size, enc_image_size, 256)
                encoded images
        """
        out = self._resnet(images)  # (batch_size, 256, 28, 28)
        self._log().debug("forward: resnet out: {}".format(out.size()))
        out = self._adaptive_pool(out)
        out = out.permute(
            0, 2, 3, 1
        )  # (batch_size, enc_image_size, enc_image_size, 256)

        self._log().debug("enc forward: final out: {}".format(out.size()))

        return out



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import json
import logging
import os
from glob import glob
from html import escape

import jsonlines
import numpy as np
import torch
import torch.utils.data
from lxml import html
from PIL import Image
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset
from tqdm import tqdm

import docling_ibm_models.tableformer.common as c
import docling_ibm_models.tableformer.settings as s
import docling_ibm_models.tableformer.utils.utils as u
from docling_ibm_models.tableformer.data_management.data_transformer import (
    DataTransformer,
)

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


class TFDataset(Dataset):
    def __init__(self, config, purpose, fixed_padding=False):
        r"""
        Parameters
        ----------
        config : Dictionary
            The input configuration file
        purpose : string
            One of s.TRAIN_PURPOSE, s.VAL_PURPOSE, s.PREDICT_PURPOSE
        fixed_padding : bool
            If False (default), the produced tag sequences will be truncated to the maximum
            actual length among the tag sequences of the batch
            If True the produced tag and cell sequences will have fixed length equal to
            max_tag_len and max_cell_len respectively.
        """
        self.cml_task = {}
        self.cml_logger = {}
        self._config = config
        self._fixed_padding = fixed_padding
        self._index = 0  # Index to the current image file
        self._max_tag_len = c.safe_get_parameter(
            config, ["preparation", "max_tag_len"], required=True
        )
        self._max_cell_len = c.safe_get_parameter(
            config, ["preparation", "max_cell_len"], required=True
        )
        self._resized_image = c.safe_get_parameter(
            config, ["dataset", "resized_image"], required=True
        )
        self.annotation = c.safe_get_parameter(
            config, ["preparation", "annotation"], required=True
        )
        self._image_normalization = config["dataset"]["image_normalization"]

        self._load_cells = c.safe_get_parameter(
            config, ["dataset", "load_cells"], required=False
        )
        self._predict_dir = c.safe_get_parameter(
            config, ["predict", "predict_dir"], required=False
        )
        self._train_bbox = c.safe_get_parameter(
            config, ["train", "bbox"], required=False
        )
        self._predict_bbox = c.safe_get_parameter(
            config, ["predict", "bbox"], required=False
        )

        self._log().debug("purpose: {}".format(purpose))
        self._log().debug("resized_image: {}".format(self._resized_image))
        self._log().debug("image_normalization: {}".format(self._image_normalization))

        # Check the type of the dataset
        dataset_type = c.safe_get_parameter(config, ["dataset", "type"])
        if dataset_type not in s.supported_datasets:
            msg = "Unsupported dataset type: " + dataset_type
            self._log().error(msg)
            raise NotImplementedError(msg)
        self._dataset_type = dataset_type

        # Check the purpose of the object
        if purpose not in [
            s.TRAIN_PURPOSE,
            s.VAL_PURPOSE,
            s.TEST_PURPOSE,
            s.PREDICT_PURPOSE,
        ]:
            msg = "Unsupported purpose: " + purpose
            self._log().error(msg)
            raise Exception(msg)
        self._purpose = purpose

        # The batch_size is grounded to 1 in case of VAL, PREDICT
        if purpose == s.TRAIN_PURPOSE:
            self._batch_size = c.safe_get_parameter(
                config, ["train", "batch_size"], required=True
            )
        else:
            self._batch_size = 1
        self._log().debug("batch_size: {}".format(self._batch_size))

        self._transformer = DataTransformer(config)
        if purpose == s.PREDICT_PURPOSE:
            self._build_predict_cache()
        else:
            self._build_cache()

        self._index = 0  # Index to the current image file
        self._ind = np.array(range(self._dataset_size))

    def set_device(self, device):
        r"""
        Set the device to be used to place the tensors when looping over the data

        Parameters
        ----------
        device : torch.device or int
            The device to do the training
        """
        self._device = device

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def __len__(self):
        return int(self._dataset_size)

    def get_batch_num(self):
        return int(np.ceil(self._dataset_size / self._batch_size))

    def _pad_tensors_from_batch(self, batch, field):
        r"""
        Retrieves data about tensor - "field" from the raw DataLoader batch list
        And pads it to maximum batch length, to be further collated in the
        custom collator function

        Parameters
        ----------
        batch : list
            The list of samples, obtained by DataLoader from the sampler
            a non-collated batch
        field: string
            Name of the tensor data-point from the sample, that has to be collated,
            and because of that has to be padded to max length of the tensor in the batch

        Returns
        -------
        batchoftensors: list
            The list of padded tensors
        """
        # Make list of selected items by field name
        list_of_items = [x[field] for x in batch]
        tagseqlens = []
        # Identify lengths of tensor "lists"
        for i, ten in enumerate(list_of_items):
            tagnum = ten[0].size()[0]
            tagseqlens.append(tagnum)
        # Get the biggest length
        maxbatchtaglen = max(tagseqlens)

        # Prepare new list with padded tensors
        batchoftensors = []
        for i, ten in enumerate(list_of_items):
            tagtensor = ten
            newtagtensor = torch.zeros(
                1, maxbatchtaglen, dtype=torch.long, device=self._device
            )
            newtagtensor[:, : tagtensor.size()[1]] = tagtensor
            batchoftensors.append(newtagtensor)

        return batchoftensors

    def bcol(self, batch):
        r"""
        Custom collate fucntion for Pytorch DataLoader, to collate items prepared
        by TFDataset into batches

        Parameters
        ----------
        batch : list
            The list of samples, obtained by DataLoader from the sampler
            a non-collated batch

        Returns
        -------
        batchoftensors: tuple
            Tuple of lists of items, properly collated into batches
        """
        collated = {}
        test_gt = {}
        if bool(batch[0]["test_gt"]):
            for x in batch:
                test_gt.update(x["test_gt"])

        if len(batch) > 1:
            # In case batch length is more than 1 we want to collate all elements in batch
            # Every element has it's own rule how to collate

            if self._load_cells:
                cells = [
                    item for sublist in [x["cells"] for x in batch] for item in sublist
                ]
            else:
                cells = []
            cell_lens = [
                item for sublist in [x["cell_lens"] for x in batch] for item in sublist
            ]
            cell_bboxes = []

            if "cell_bboxes" in batch[0] and batch[0]["cell_bboxes"]:
                for x in batch:
                    cell_bboxes.append(
                        {
                            "boxes": x["cell_bboxes"][0][0],
                            "labels": x["cell_bboxes"][0][1],
                        }
                    )

            samples = [
                item for sublist in [x["samples"] for x in batch] for item in sublist
            ]

            # Sequences of tags have to be padded and then collated:
            batchoftags = self._pad_tensors_from_batch(batch, "tags")
            tags = pad_sequence(batchoftags, batch_first=True)
            tags = torch.squeeze(tags, 1)

            tag_lens = pad_sequence([x["tag_lens"] for x in batch], batch_first=True)
            tag_lens = torch.squeeze(tag_lens, 1)
            num_cells = pad_sequence([x["num_cells"] for x in batch], batch_first=True)
            num_cells = torch.squeeze(num_cells, 1)

            collated = (
                torch.cat([x["imgs"] for x in batch], dim=0),
                tags,
                tag_lens,
                num_cells,
                cells,
                cell_lens,
                cell_bboxes,
                samples,
                test_gt,
            )
        else:
            # In case batch length is 1 we just formulate the expected output
            cell_bboxes = None
            if "cell_bboxes" in batch[0] and batch[0]["cell_bboxes"]:
                cell_bboxes = {
                    "boxes": batch[0]["cell_bboxes"][0][0],
                    "labels": batch[0]["cell_bboxes"][0][1],
                }

            collated = (
                batch[0]["imgs"],
                batch[0]["tags"],
                batch[0]["tag_lens"],
                batch[0]["num_cells"],
                batch[0]["cells"],
                batch[0]["cell_lens"],
                cell_bboxes,
                batch[0]["samples"],
                test_gt,
            )
        return collated

    def __getitem__(self, idx):
        r"""
        Retrieve data by the specific index from the cache
        Required for Pytorch DataSampler, and to be used together with DataLoader
        Depending on the "purpose" different returned objects can be None.

        Returns
        (All data points presented in a dictionary, each wrapped into a list,
        for easier batching and collating later on)
        -------
        imgs : tensor (batch_size, image_channels, resized_image, resized_image)
            Batch with the rescaled images
        tags : The tags of the images. The object is one of:
            None : If purpose is not "train"
            (batch_size, max_tag_len + 2) : If purpose is "train" and "fixed_padding" is true.
                                            The +2 over the max_tag_len is for the <start> <stop>
            (batch_size, batch_max_tag_len) : If purpose is "train" and "fixed_padding" is false,
                                              where "batch_max_tag_len" is the max length of the
                                              tags in batch
        tag_lens : The real length of the tags per image in the batch. The object is one of:
            None : If purpose is not "train"
            (batch_size, 1) : If purpose is "train"
        num_cells : The number of cells per image in the batch. The object is one of:
            None : If purpose is not "train"
            (batch_size, 1) : If purpose is "train"
        cells : The cell tags for the images in the batch. The object is one of:
            None : If purpose is not "train"
            list with LongTensor: If purpose is "train"
        cell_lens : The length of the cell tags per image in the batch. The object is one of:
            None : If purpose is not "train"
            list with LongTensor: If purpose is "train"
        cell_bboxes : The transformed (rescaled, padded, etc) bboxes of the cells for all images in
                      batch. Each list is a bbox in the format [xc, yc, w, h] where xc, yc are the
                      coords of the center, w, h the width and height of the bbox and all are
                      normalized to the scaled size of the image. The object is one of:
            None : If purpose is not "train"
            list of lists: If purpose is "train"
        samples : list of string
            The filenames in the batch
        val_gt : The ground turth raw attributes for the validation split. The object is one of:
            None : If purpose is not "val"
            dictionary : If purpose is "val"
        """

        # Check if index out of bounds
        if idx >= self._dataset_size:
            return None

        # Move current _index to requested idx
        # self._index = idx

        # Images, __getitem__ provides only 1 image for specified index
        # (batch_size, image_channels, resized_image, resized_image)
        imgs = torch.zeros(
            1, self._img_ch, self._resized_image, self._resized_image, dtype=torch.float
        ).to(self._device)

        # Initialize all output objects to None
        # Depending on the "purpose" some of them will be populated
        tags = None
        tag_lens = None
        num_cells = None
        cells = None
        cell_lens = None
        cell_bboxes = None

        val_tags = None
        val_tag_lens = None
        val_num_cells = None
        val_cells = None
        val_cell_lens = None
        val_cell_bboxes = None

        test_gt = {}
        test_tags = None
        test_tag_lens = None

        # Train specific output
        if self._purpose == s.TRAIN_PURPOSE:
            tag_len = self._taglens[idx]
            fixed_tag_len = self._max_tag_len + 2  # <start>...<end>
            if self._fixed_padding:
                tags = torch.zeros(
                    1, fixed_tag_len, dtype=torch.long, device=self._device
                )
            else:
                tags = torch.zeros(1, tag_len, dtype=torch.long, device=self._device)

            tag_lens = torch.zeros(1, 1, dtype=torch.long).to(self._device)
            num_cells = torch.zeros(1, 1, dtype=torch.long).to(self._device)
            cells = []
            cell_lens = []
            cell_bboxes = []
        # val specific output
        elif self._purpose == s.VAL_PURPOSE:
            val_tag_len = self._val_taglens[idx]
            val_fixed_tag_len = self._max_tag_len + 2  # <start>...<end>
            if self._fixed_padding:
                val_tags = torch.zeros(
                    1, val_fixed_tag_len, dtype=torch.long, device=self._device
                )
            else:
                val_tags = torch.zeros(
                    1, val_tag_len, dtype=torch.long, device=self._device
                )

            val_tag_lens = torch.zeros(1, 1, dtype=torch.long).to(self._device)
            val_num_cells = torch.zeros(1, 1, dtype=torch.long).to(self._device)
            val_cells = []
            val_cell_lens = []
            val_cell_bboxes = []

        elif self._purpose == s.TEST_PURPOSE:
            if len(self._test_taglens) > 0:
                # Dictionary with the raw attributes for the groundtruth. Keys are the filenames
                test_gt = {}
                tag_len = self._test_taglens[idx]
                fixed_tag_len = self._max_tag_len + 2  # <start>...<end>
                if self._fixed_padding:
                    test_tags = torch.zeros(
                        1, fixed_tag_len, dtype=torch.long, device=self._device
                    )
                else:
                    test_tags = torch.zeros(
                        1, tag_len, dtype=torch.long, device=self._device
                    )
                test_tag_lens = torch.zeros(1, 1, dtype=torch.long).to(self._device)
                cells = []
                cell_lens = []
                cell_bboxes = []

        sample = self._image_fns[idx]
        # Rescale/convert the image and bboxes
        bboxes = self._bboxes[sample]
        sample_fn = self._get_image_path(sample)

        if not self._table_bboxes:
            table_bbox = None
        else:
            if sample in self._table_bboxes:
                table_bbox = self._table_bboxes[sample]
            else:
                table_bbox = None
        scaled_img, scaled_bboxes = self._transformer.sample_preprocessor(
            sample_fn, bboxes, self._purpose, table_bbox
        )

        imgs[0] = scaled_img.to(self._device)

        # Train specific output
        if self._purpose == s.TRAIN_PURPOSE:
            # Remove the padding from tags and cells
            if self._fixed_padding:
                tags[0] = torch.LongTensor(self._tags[idx]).to(self._device)
            else:
                tags[0] = torch.LongTensor(self._tags[idx][:tag_len]).to(self._device)

            tag_lens[0] = torch.LongTensor([self._taglens[idx]]).to(self._device)
            num_cells[0] = len(self._cell_lens[idx])

            if len(self._cell_lens[idx]) > 0:
                sample_max_cell_len = max(self._cell_lens[idx])
            else:
                sample_max_cell_len = 0

            if self._load_cells:
                image_trimmed_cells = [
                    self._cells[idx][x][0:sample_max_cell_len]
                    for x in range(0, len(self._cells[idx]))
                ]
                cells.append(torch.LongTensor(image_trimmed_cells).to(self._device))

            cell_lens.append(torch.LongTensor(self._cell_lens[idx]).to(self._device))
            if self._train_bbox:

                cell_bboxes.append(
                    [
                        torch.from_numpy(
                            np.array(scaled_bboxes, dtype=np.float32)[:, :4]
                        ).to(self._device),
                        torch.from_numpy(
                            np.array(scaled_bboxes, dtype=np.compat.long)[:, -1]
                        ).to(self._device),
                    ]
                )

        elif self._purpose == s.VAL_PURPOSE:
            # Remove the padding from tags and cells
            if self._fixed_padding:
                val_tags[0] = torch.LongTensor(self._val_tags[idx]).to(self._device)
            else:
                val_tags[0] = torch.LongTensor(self._val_tags[idx][:val_tag_len]).to(
                    self._device
                )

            val_tag_lens[0] = torch.LongTensor([self._val_taglens[idx]]).to(
                self._device
            )
            val_num_cells[0] = len(self._val_cell_lens[idx])

            if len(self._val_cell_lens[idx]) > 0:
                sample_max_cell_len = max(self._val_cell_lens[idx])
            else:
                sample_max_cell_len = 0

            if self._load_cells:
                val_image_trimmed_cells = [
                    self._val_cells[idx][x][0:sample_max_cell_len]
                    for x in range(0, len(self._cells[idx]))
                ]
                val_cells.append(
                    torch.LongTensor(val_image_trimmed_cells).to(self._device)
                )

            val_cell_lens.append(
                torch.LongTensor(self._val_cell_lens[idx]).to(self._device)
            )
            if self._train_bbox:
                val_cell_bboxes.append(
                    [
                        torch.from_numpy(
                            np.array(scaled_bboxes, dtype=np.float32)[:, :4]
                        ).to(self._device),
                        torch.from_numpy(
                            np.array(scaled_bboxes, dtype=np.compat.long)[:, -1]
                        ).to(self._device),
                    ]
                )
        # val specific output
        elif self._purpose == s.TEST_PURPOSE:
            if test_gt is not None:
                test_gt[sample] = self._test[sample]

                # Remove the padding from tags and cells
                if self._fixed_padding:
                    test_tags[0] = torch.LongTensor(self._test_tags[idx]).to(
                        self._device
                    )
                else:
                    test_tags[0] = torch.LongTensor(self._test_tags[idx][:tag_len]).to(
                        self._device
                    )
                test_tag_lens[0] = torch.LongTensor([self._test_taglens[idx]]).to(
                    self._device
                )
                if self._predict_bbox:
                    cell_bboxes.append(
                        [
                            torch.from_numpy(
                                np.array(scaled_bboxes, dtype=np.float32)[:, :4]
                            ).to(self._device),
                            torch.from_numpy(
                                np.array(scaled_bboxes, dtype=np.compat.long)[:, -1]
                            ).to(self._device),
                        ]
                    )

        output = {}

        # Samples is a list with the given image filename
        samples = [self._image_fns[idx]]
        # All data points presented in a dictionary, each wrapped into a list,
        # for easier batching and collating later on
        if self._purpose == s.TRAIN_PURPOSE:
            output["imgs"] = imgs
            output["tags"] = tags
            output["tag_lens"] = tag_lens
            output["num_cells"] = num_cells
            output["cells"] = cells
            output["cell_lens"] = cell_lens
            output["cell_bboxes"] = cell_bboxes
            output["samples"] = samples
            output["test_gt"] = test_gt
        elif self._purpose == s.VAL_PURPOSE:
            output["imgs"] = imgs
            output["tags"] = val_tags
            output["tag_lens"] = val_tag_lens
            output["num_cells"] = val_num_cells
            output["cells"] = val_cells
            output["cell_lens"] = val_cell_lens
            output["cell_bboxes"] = val_cell_bboxes
            output["samples"] = samples
            output["test_gt"] = test_gt
        elif self._purpose == s.TEST_PURPOSE:
            output["imgs"] = imgs
            output["tags"] = test_tags
            output["tag_lens"] = test_tag_lens
            output["num_cells"] = num_cells
            output["cells"] = cells
            output["cell_lens"] = cell_lens
            output["cell_bboxes"] = cell_bboxes
            output["samples"] = samples
            output["test_gt"] = test_gt
        else:
            output["imgs"] = imgs
            output["tags"] = tags
            output["tag_lens"] = tag_lens
            output["num_cells"] = num_cells
            output["cells"] = cells
            output["cell_lens"] = cell_lens
            output["cell_bboxes"] = cell_bboxes
            output["samples"] = samples
            output["test_gt"] = test_gt
        return output

    def get_batch_size(self):
        r"""
        Return the actual batch_size
        """
        return self._batch_size

    def reset(self):
        self._index = 0

    def __iter__(self):
        return self

    def is_valid(self, img, config):
        max_tag_len = config["preparation"]["max_tag_len"]
        max_cell_len = config["preparation"]["max_cell_len"]
        check_limits = True
        if "check_limits" in config["preparation"]:
            check_limits = config["preparation"]["check_limits"]
        if check_limits:
            if len(img["html"]["structure"]["tokens"]) > max_tag_len:
                self._log().debug(
                    "========================================= TAG LEN REJECTED"
                )
                self._log().debug("File name: {}".format(img["filename"]))
                tokens_len = len(img["html"]["structure"]["tokens"])
                self._log().debug("Structure token len: {}".format(tokens_len))
                self._log().debug("More than max_tag_len: {}".format(max_tag_len))
                self._log().debug(
                    "=========================================================="
                )
                return False
            for cell in img["html"]["cells"]:
                if len(cell["tokens"]) > max_cell_len:
                    self._log().debug(
                        "======================================== CELL LEN REJECTED"
                    )
                    self._log().debug("File name: {}".format(img["filename"]))
                    self._log().debug("Cell len: {}".format(len(cell["tokens"])))
                    self._log().debug("More than max_cell_len: {}".format(max_cell_len))
                    self._log().debug(
                        "=========================================================="
                    )
                    return False
        self.raw_data_dir = config["preparation"]["raw_data_dir"]
        with Image.open(
            os.path.join(self.raw_data_dir, img["split"], img["filename"])
        ) as im:
            max_image_size = config["preparation"]["max_image_size"]
            if im.width > max_image_size or im.height > max_image_size:
                # IMG SIZE REJECTED
                return False
        return True

    def __next__(self):
        r"""
        Get the next batch or raise the StopIteration

        In order to have the batch size fixed also in the last iteration, we wrap over the dataset
        and repeat some of the first elements.

        Depending on the "purpose" different returned objects can be None.

        Returns
        -------
        imgs : tensor (batch_size, image_channels, resized_image, resized_image)
            Batch with the rescaled images
        tags : The tags of the images. The object is one of:
            None : If purpose is not "train"
            (batch_size, max_tag_len + 2) : If purpose is "train" and "fixed_padding" is true.
                                            The +2 over the max_tag_len is for the <start> <stop>
            (batch_size, batch_max_tag_len) : If purpose is "train" and "fixed_padding" is false,
                                              where "batch_max_tag_len" is the max length of the
                                              tags in batch
        tag_lens : The real length of the tags per image in the batch. The object is one of:
            None : If purpose is not "train"
            (batch_size, 1) : If purpose is "train"
        num_cells : The number of cells per image in the batch. The object is one of:
            None : If purpose is not "train"
            (batch_size, 1) : If purpose is "train"
        cells : The cell tags for the images in the batch. The object is one of:
            None : If purpose is not "train"
            list with LongTensor: If purpose is "train"
        cell_lens : The length of the cell tags per image in the batch. The object is one of:
            None : If purpose is not "train"
            list with LongTensor: If purpose is "train"
        cell_bboxes : The transformed (rescaled, padded, etc) bboxes of the cells for all images in
                      batch. Each list is a bbox in the format [xc, yc, w, h] where xc, yc are the
                      coords of the center, w, h the width and height of the bbox and all are
                      normalized to the scaled size of the image. The object is one of:
            None : If purpose is not "train"
            list of lists: If purpose is "train"
        samples : list of string
            The filenames in the batch
        val_gt : The ground turth raw attributes for the validation split. The object is one of:
            None : If purpose is not "val"
            dictionary : If purpose is "val"
        """

        if self._index >= self._dataset_size:
            raise StopIteration()

        # Compute the next sample
        if (
            self._dataset_size - self._index >= self._batch_size
        ):  # Full batch_size sample
            step = self._batch_size
            sample_indices = self._ind[self._index : self._index + step]
        else:
            # skip last batch
            raise StopIteration()

        batch = []
        # Loop over the batch indices and collect items for the batch
        for i, idx in enumerate(sample_indices):
            item = self.__getitem__(idx)
            batch.append(item)
        # Collate batch
        output = self.bcol(batch)
        self._index += step

        return output

    def shuffle(self):
        r"""
        Shuffle the training images
        This takes place only in case of training, otherwise it just returns

        Output: True in case a shuffling took place, False otherwise
        """
        if self._purpose != s.TRAIN_PURPOSE:
            return False

        # image_fns_np = np.asarray(self._image_fns)
        # To get a deterministic random shuffle, we need to seed our random
        # with a deterministic seed (int)
        np.random.seed(42)
        # Then shuffle after seeding
        self._ind = np.random.permutation(self._dataset_size)
        self._index = 0
        return True

    def get_init_data(self):
        r"""
        Create a dictionary with all kind of initialization data necessary for all models.
        This data should not be served by the __next__ method.
        """
        init_data = {"word_map": self._word_map, "statistics": self._statistics}
        return init_data

    def _get_image_path(self, img_fn):
        r"""
        Get the full image path out of the image file name
        """
        if self._dataset_type == "TF_prepared":
            if self._purpose == s.TRAIN_PURPOSE:
                full_fn = os.path.join(self._raw_data_dir, "train", img_fn)
            elif self._purpose == s.VAL_PURPOSE:
                full_fn = os.path.join(self._raw_data_dir, "val", img_fn)
            elif self._purpose == s.TEST_PURPOSE:
                full_fn = os.path.join(self._raw_data_dir, "test", img_fn)
            else:
                full_fn = os.path.join(self._raw_data_dir, img_fn)

        if full_fn is None or not os.path.isfile(full_fn):
            self._log().error("File not found: " + full_fn)
            return None

        return full_fn

    def format_html(self, img):
        r"""
        Formats HTML code from tokenized annotation of img
        """
        tag_len = len(img["html"]["structure"]["tokens"])
        if self._load_cells:
            cell_len_max = max([len(c["tokens"]) for c in img["html"]["cells"]])
        else:
            cell_len_max = 0

        HTML = img["html"]["structure"]["tokens"].copy()
        to_insert = [i for i, tag in enumerate(HTML) if tag in ("<td>", ">")]

        if self._load_cells:
            for i, cell in zip(to_insert[::-1], img["html"]["cells"][::-1]):
                if cell:
                    cell = "".join(
                        [
                            escape(token) if len(token) == 1 else token
                            for token in cell["tokens"]
                        ]
                    )
                    HTML.insert(i + 1, cell)

        HTML = "<html><body><table>%s</table></body></html>" % "".join(HTML)
        root = html.fromstring(HTML)
        if self._predict_bbox:
            for td, cell in zip(root.iter("td"), img["html"]["cells"]):
                if "bbox" in cell:
                    bbox = cell["bbox"]
                    td.attrib["x"] = str(bbox[0])
                    td.attrib["y"] = str(bbox[1])
                    td.attrib["width"] = str(bbox[2] - bbox[0])
                    td.attrib["height"] = str(bbox[3] - bbox[1])
        HTML = html.tostring(root, encoding="utf-8").decode()
        return HTML, tag_len, cell_len_max

    def _build_predict_cache(self):
        r"""
        populate cache with image file names that need to be predicted
        """
        self._prepared_data_dir = c.safe_get_parameter(
            self._config, ["dataset", "prepared_data_dir"], required=False
        )
        self._data_name = c.safe_get_parameter(
            self._config, ["dataset", "name"], required=True
        )

        if self._prepared_data_dir is None:

            self._statistics = c.safe_get_parameter(
                self._config, ["dataset", "image_normalization"], required=True
            )

            self._word_map = c.safe_get_parameter(
                self._config, ["dataset_wordmap"], required=True
            )

        else:
            # Load statistics
            statistics_fn = c.get_prepared_data_filename("STATISTICS", self._data_name)
            with open(os.path.join(self._prepared_data_dir, statistics_fn), "r") as f:
                self._log().debug("Load statistics from: {}".format(statistics_fn))
                self._statistics = json.load(f)

            # Load word_map
            word_map_fn = c.get_prepared_data_filename("WORDMAP", self._data_name)
            with open(os.path.join(self._prepared_data_dir, word_map_fn), "r") as f:
                self._log().debug("Load WORDMAP from: {}".format(word_map_fn))
                self._word_map = json.load(f)

        # Get Image File Names for Prediction
        self._image_fns = []
        self._bboxes = {}
        self._bboxes_table = {}
        self._raw_data_dir = self._predict_dir
        self._table_bboxes = {}

        if self._predict_dir[-1] != "/":
            self._predict_dir += "/"
        for path in list(glob(self._predict_dir + "**/*.png", recursive=True)):
            filename = os.path.basename(path)
            self._image_fns.append(filename)
            self._log().info("Image found: {}".format(filename))
            self._bboxes[filename] = []

        # Get size of a dataset to predict
        self._dataset_size = len(self._image_fns)

        # Get the number of image channels
        self._log().info(
            "To test load... {}".format(self._predict_dir + self._image_fns[0])
        )
        img = u.load_image(self._predict_dir + self._image_fns[0])
        if img is None:
            msg = "Cannot load image"
            self._log().error(msg)
            raise Exception(msg)
        self._img_ch = img.shape[0]

    def _build_cache(self):
        r"""
        Cache with small data
        """
        all_bboxes = {}  # Keep original bboxes for all images
        table_bboxes = {}
        self._log().info("Building the cache...")
        self._raw_data_dir = c.safe_get_parameter(
            self._config, ["dataset", "raw_data_dir"], required=True
        )
        self._prepared_data_dir = c.safe_get_parameter(
            self._config, ["dataset", "prepared_data_dir"], required=False
        )

        self._data_name = c.safe_get_parameter(
            self._config, ["dataset", "name"], required=True
        )

        if self._prepared_data_dir is None:

            self._statistics = c.safe_get_parameter(
                self._config, ["dataset", "image_normalization"], required=True
            )

            self._word_map = c.safe_get_parameter(
                self._config, ["dataset_wordmap"], required=True
            )

        else:
            # Load statistics
            statistics_fn = c.get_prepared_data_filename("STATISTICS", self._data_name)
            with open(os.path.join(self._prepared_data_dir, statistics_fn), "r") as f:
                self._log().debug("Load statistics from: {}".format(statistics_fn))
                self._statistics = json.load(f)

            # Load word_map
            word_map_fn = c.get_prepared_data_filename("WORDMAP", self._data_name)
            with open(os.path.join(self._prepared_data_dir, word_map_fn), "r") as f:
                self._log().debug("Load WORDMAP from: {}".format(word_map_fn))
                self._word_map = json.load(f)

        word_map_cell = self._word_map["word_map_cell"]
        word_map_tag = self._word_map["word_map_tag"]
        # Read image paths and captions for each image
        train_image_paths = []
        train_images = []
        val_image_paths = []
        val_images = []
        test_image_paths = []
        test_images = []
        predict_images = []

        train_image_tags = (
            []
        )  # List of list of structure tokens for each image in the train set
        train_image_cells = []
        train_image_cells_len = []

        val_image_tags = (
            []
        )  # List of list of structure tokens for each image in the train set
        val_image_cells = []
        val_image_cells_len = []

        test_image_tags = []
        test_gt = dict()

        invalid_files = 0
        total_files = 0

        self._log().debug(
            "Create lists with image filenames per split, train tags/cells and GT"
        )
        with jsonlines.open(self.annotation, "r") as reader:
            for img in tqdm(reader):
                total_files += 1
                img_filename = img["filename"]
                path = os.path.join(self._raw_data_dir, img["split"], img_filename)

                # Keep bboxes for all images
                all_cell_bboxes = []
                for cell in img["html"]["cells"]:
                    if "bbox" not in cell:
                        continue
                    all_cell_bboxes.append(cell["bbox"])
                all_bboxes[img_filename] = all_cell_bboxes

                # if dataset does include bbox for the table itself
                if "table_bbox" in img:
                    table_bboxes[img_filename] = img["table_bbox"]
                if img["split"] == "train":
                    if self._purpose == s.TRAIN_PURPOSE:
                        # Skip invalid images
                        if not self.is_valid(img, self._config):
                            invalid_files += 1
                            continue
                        tags = []
                        cells = []
                        cell_lens = []
                        tags.append(img["html"]["structure"]["tokens"])

                        if self._load_cells:
                            for cell in img["html"]["cells"]:
                                cells.append(cell["tokens"])
                                cell_lens.append(len(cell["tokens"]) + 2)
                        else:
                            for cell in img["html"]["cells"]:
                                cell_lens.append(len(cell["tokens"]) + 2)

                        train_images.append(img_filename)
                        train_image_paths.append(path)
                        train_image_tags.append(tags)
                        train_image_cells.append(cells)
                        train_image_cells_len.append(cell_lens)
                if img["split"] == "val":
                    if self._purpose == s.VAL_PURPOSE:
                        # Skip invalid images
                        if not self.is_valid(img, self._config):
                            invalid_files += 1
                            continue

                        val_tags = []
                        val_cells = []
                        val_cell_lens = []
                        val_tags.append(img["html"]["structure"]["tokens"])

                        if self._load_cells:
                            for cell in img["html"]["cells"]:
                                val_cells.append(cell["tokens"])
                                val_cell_lens.append(len(cell["tokens"]) + 2)
                        else:
                            for cell in img["html"]["cells"]:
                                val_cell_lens.append(len(cell["tokens"]) + 2)

                        with Image.open(path) as im:
                            HTML, tag_len, cell_len_max = self.format_html(img)
                            lt1 = [">", "lcel", "ucel", "xcel"]
                            lt2 = img["html"]["structure"]["tokens"]
                            tcheck = any(item in lt1 for item in lt2)
                            if tcheck:
                                gtt = "complex"
                            else:
                                gtt = "simple"
                            test_gt[img_filename] = {
                                "html": HTML,
                                "tag_len": tag_len,
                                "cell_len_max": cell_len_max,
                                "width": im.width,
                                "height": im.height,
                                "type": gtt,
                                "html_tags": img["html"]["structure"]["tokens"],
                            }

                        val_images.append(img_filename)
                        val_image_paths.append(path)
                        val_image_tags.append(val_tags)
                        val_image_cells.append(val_cells)
                        val_image_cells_len.append(val_cell_lens)

                elif img["split"] == "test":
                    if self._purpose == s.TEST_PURPOSE:
                        # Skip invalid images
                        if not self.is_valid(img, self._config):
                            invalid_files += 1
                            continue

                        with Image.open(path) as im:
                            HTML, tag_len, cell_len_max = self.format_html(img)
                            lt1 = [">", "lcel", "ucel", "xcel"]
                            lt2 = img["html"]["structure"]["tokens"]
                            tcheck = any(item in lt1 for item in lt2)
                            if tcheck:
                                gtt = "complex"
                            else:
                                gtt = "simple"
                            test_gt[img_filename] = {
                                "html": HTML,
                                "tag_len": tag_len,
                                "cell_len_max": cell_len_max,
                                "width": im.width,
                                "height": im.height,
                                "type": gtt,
                                "html_tags": img["html"]["structure"]["tokens"],
                            }
                        test_images.append(img_filename)

                        test_tags = []
                        test_tags.append(img["html"]["structure"]["tokens"])
                        test_image_paths.append(path)
                        test_image_tags.append(test_tags)
                else:
                    if self._purpose == s.PREDICT_PURPOSE:
                        predict_images.append(img_filename)

        image_fns = {
            s.TRAIN_SPLIT: train_images,
            s.VAL_SPLIT: val_images,
            s.TEST_SPLIT: test_images,
        }

        self._log().debug("Keep the split data pointed by the purpose")
        # Images
        # Filter out the images for the particual split
        self._image_fns = image_fns[self._purpose]
        self._dataset_size = len(self._image_fns)
        assert len(self._image_fns) > 0, "Empty image split: " + self._purpose

        # Get the number of image channels
        img = u.load_image(self._get_image_path(self._image_fns[0]))
        if img is None:
            msg = "Cannot load image"
            self._log().error(msg)
            raise Exception(msg)
        self._img_ch = img.shape[0]

        # img_name -> list of bboxes, each bbox is a list with x1y1x2y2
        split_bboxes = {}
        img_names = set(self._image_fns)  # Set will speed up search
        for img_name, bbox in all_bboxes.items():
            if img_name not in img_names:
                continue
            if img_name not in split_bboxes:
                split_bboxes[img_name] = []
            # we should use extend not append, otherwise we get list within list
            split_bboxes[img_name].extend(bbox)
        self._bboxes = split_bboxes
        self._table_bboxes = table_bboxes
        # -------------------------------------------------------------------------------
        # Train specific
        # -------------------------------------------------------------------------------
        # Compute encoded tags and cells
        enc_tags = []
        tag_lens = []
        enc_cells = []
        cell_lens = []

        val_enc_tags = []
        val_tag_lens = []
        val_enc_cells = []
        val_cell_lens = []

        test_enc_tags = []
        test_tag_lens = []

        # Based on the "purpose"
        if self._purpose == s.TRAIN_PURPOSE:
            self._log().debug("Convert train tags and cell tags to indices")
            for i, path in enumerate(tqdm(train_image_paths)):
                for tag in train_image_tags[i]:
                    # Encode tags
                    # Notice that at this point we don't have images longer than max_tag_length
                    # The same happens with the cell tokens
                    enc_tag = (
                        [word_map_tag["<start>"]]
                        + [
                            word_map_tag.get(word, word_map_tag["<unk>"])
                            for word in tag
                        ]
                        + [word_map_tag["<end>"]]
                        + [word_map_tag["<pad>"]] * (self._max_tag_len - len(tag))
                    )
                    # Find caption lengths
                    tag_len = len(tag) + 2

                    enc_tags.append(enc_tag)
                    tag_lens.append(tag_len)

                enc_cell_seq = []
                cell_seq_len = []

                if self._load_cells:
                    for cell in train_image_cells[i]:
                        # Encode captions
                        enc_cell = (
                            [word_map_cell["<start>"]]
                            + [
                                word_map_cell.get(word, word_map_cell["<unk>"])
                                for word in cell
                            ]
                            + [word_map_cell["<end>"]]
                            + [word_map_cell["<pad>"]]
                            * (self._max_cell_len - len(cell))
                        )
                        enc_cell_seq.append(enc_cell)
                        # Find caption lengths
                        cell_len = len(cell) + 2
                        cell_seq_len.append(cell_len)
                else:
                    for cell in train_image_cells_len[i]:
                        cell_seq_len.append(cell)
                enc_cells.append(enc_cell_seq)
                cell_lens.append(cell_seq_len)

        if self._purpose == s.VAL_PURPOSE:
            self._log().debug("Convert train tags and cell tags to indices")
            for i, path in enumerate(tqdm(val_image_paths)):
                for tag in val_image_tags[i]:
                    # Encode tags
                    # Notice that at this point we don't have images longer than max_tag_length
                    # The same happens with the cell tokens
                    val_enc_tag = (
                        [word_map_tag["<start>"]]
                        + [
                            word_map_tag.get(word, word_map_tag["<unk>"])
                            for word in tag
                        ]
                        + [word_map_tag["<end>"]]
                        + [word_map_tag["<pad>"]] * (self._max_tag_len - len(tag))
                    )
                    # Find caption lengths
                    val_tag_len = len(tag) + 2

                    val_enc_tags.append(val_enc_tag)
                    val_tag_lens.append(val_tag_len)

                val_enc_cell_seq = []
                val_cell_seq_len = []

                if self._load_cells:
                    for cell in val_image_cells[i]:
                        # Encode captions
                        val_enc_cell = (
                            [word_map_cell["<start>"]]
                            + [
                                word_map_cell.get(word, word_map_cell["<unk>"])
                                for word in cell
                            ]
                            + [word_map_cell["<end>"]]
                            + [word_map_cell["<pad>"]]
                            * (self._max_cell_len - len(cell))
                        )
                        val_enc_cell_seq.append(val_enc_cell)

                    # Find caption lengths
                    cell_len = len(cell) + 2
                    val_cell_seq_len.append(cell_len)
                else:
                    for cell in val_image_cells_len[i]:
                        val_cell_seq_len.append(cell)
                val_enc_cells.append(val_enc_cell_seq)
                val_cell_lens.append(val_cell_seq_len)

        if self._purpose == s.TEST_PURPOSE:
            self._log().debug("Convert val tags to indices")
            for i, path in enumerate(tqdm(test_image_paths)):
                for tag in test_image_tags[i]:
                    # Encode tags
                    # Notice that at this point we don't have images longer than max_tag_length
                    # The same happens with the cell tokens
                    test_enc_tag = (
                        [word_map_tag["<start>"]]
                        + [
                            word_map_tag.get(word, word_map_tag["<unk>"])
                            for word in tag
                        ]
                        + [word_map_tag["<end>"]]
                        + [word_map_tag["<pad>"]] * (self._max_tag_len - len(tag))
                    )

                    # Find caption lengths
                    test_tag_len = len(tag) + 2

                    test_enc_tags.append(test_enc_tag)
                    test_tag_lens.append(test_tag_len)

        self._tags = enc_tags
        self._taglens = tag_lens
        self._cells = enc_cells
        self._cell_lens = cell_lens

        # -------------------------------------------------------------------------------
        # val specific
        # -------------------------------------------------------------------------------
        self._val_tags = val_enc_tags
        self._val_taglens = val_tag_lens
        self._val_cells = val_enc_cells
        self._val_cell_lens = val_cell_lens
        # -------------------------------------------------------------------------------
        # test / evaluation specific
        # -------------------------------------------------------------------------------
        self._test = test_gt
        self._test_tags = test_enc_tags
        self._test_taglens = test_tag_lens



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging

import torch
import torch.nn as nn

import docling_ibm_models.tableformer.settings as s
import docling_ibm_models.tableformer.utils.utils as u

# from scipy.optimize import linear_sum_assignment

LOG_LEVEL = logging.INFO


class CellAttention(nn.Module):
    """
    Attention Network.
    """

    def __init__(self, encoder_dim, tag_decoder_dim, language_dim, attention_dim):
        """
        :param encoder_dim: feature size of encoded images
        :param tag_decoder_dim: size of tag decoder's RNN
        :param language_dim: size of language model's RNN
        :param attention_dim: size of the attention network
        """
        super(CellAttention, self).__init__()
        # linear layer to transform encoded image
        self._encoder_att = nn.Linear(encoder_dim, attention_dim)
        # linear layer to transform tag decoder output
        self._tag_decoder_att = nn.Linear(tag_decoder_dim, attention_dim)
        # linear layer to transform language models output
        self._language_att = nn.Linear(language_dim, attention_dim)
        # linear layer to calculate values to be softmax-ed
        self._full_att = nn.Linear(attention_dim, 1)
        self._relu = nn.ReLU()
        self._softmax = nn.Softmax(dim=1)  # softmax layer to calculate weights

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def forward(self, encoder_out, decoder_hidden, language_out):
        """
        Forward propagation.
        :param encoder_out: encoded images, a tensor of dimension (1, num_pixels, encoder_dim)
        :param decoder_hidden: tag decoder output, a tensor of dimension [(num_cells,
                               tag_decoder_dim)]
        :param language_out: language model output, a tensor of dimension (num_cells,
                               language_dim)
        :return: attention weighted encoding, weights
        """
        att1 = self._encoder_att(encoder_out)  # (1, num_pixels, attention_dim)
        att2 = self._tag_decoder_att(decoder_hidden)  # (num_cells, tag_decoder_dim)
        att3 = self._language_att(language_out)  # (num_cells, attention_dim)
        att = self._full_att(
            self._relu(att1 + att2.unsqueeze(1) + att3.unsqueeze(1))
        ).squeeze(2)
        alpha = self._softmax(att)  # (num_cells, num_pixels)
        # (num_cells, encoder_dim)
        attention_weighted_encoding = (encoder_out * alpha.unsqueeze(2)).sum(dim=1)
        return attention_weighted_encoding, alpha


class BBoxDecoder(nn.Module):
    """
    CellDecoder generates cell content
    """

    def __init__(
        self,
        device,
        attention_dim,
        embed_dim,
        tag_decoder_dim,
        decoder_dim,
        num_classes,
        encoder_dim=512,
        dropout=0.5,
        cnn_layer_stride=1,
    ):
        """
        :param attention_dim: size of attention network
        :param embed_dim: embedding size
        :param tag_decoder_dim: size of tag decoder's RNN
        :param decoder_dim: size of decoder's RNN
        :param vocab_size: size of vocabulary
        :param encoder_dim: feature size of encoded images
        :param dropout: dropout
        :param mini_batch_size: batch size of cells to reduce GPU memory usage
        """
        super(BBoxDecoder, self).__init__()
        self._device = device
        self._encoder_dim = encoder_dim
        self._attention_dim = attention_dim
        self._embed_dim = embed_dim
        self._decoder_dim = decoder_dim
        self._dropout = dropout
        self._num_classes = num_classes

        if cnn_layer_stride is not None:
            self._input_filter = u.resnet_block(stride=cnn_layer_stride)
        # attention network
        self._attention = CellAttention(
            encoder_dim, tag_decoder_dim, decoder_dim, attention_dim
        )
        # decoder LSTMCell
        self._init_h = nn.Linear(encoder_dim, decoder_dim)

        # linear layer to create a sigmoid-activated gate
        self._f_beta = nn.Linear(decoder_dim, encoder_dim)
        self._sigmoid = nn.Sigmoid()
        self._dropout = nn.Dropout(p=self._dropout)
        self._class_embed = nn.Linear(512, self._num_classes + 1)
        self._bbox_embed = u.MLP(512, 256, 4, 3)

    def _init_hidden_state(self, encoder_out, batch_size):
        mean_encoder_out = encoder_out.mean(dim=1)
        h = self._init_h(mean_encoder_out).expand(batch_size, -1)
        return h

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def inference(self, encoder_out, tag_H):
        """
        Inference on test images with beam search
        """
        if hasattr(self, "_input_filter"):
            encoder_out = self._input_filter(encoder_out.permute(0, 3, 1, 2)).permute(
                0, 2, 3, 1
            )

        encoder_dim = encoder_out.size(3)

        # Flatten encoding (1, num_pixels, encoder_dim)
        encoder_out = encoder_out.view(1, -1, encoder_dim)

        num_cells = len(tag_H)
        predictions_bboxes = []
        predictions_classes = []

        for c_id in range(num_cells):
            # Start decoding
            h = self._init_hidden_state(encoder_out, 1)
            cell_tag_H = tag_H[c_id]
            awe, _ = self._attention(encoder_out, cell_tag_H, h)
            gate = self._sigmoid(self._f_beta(h))
            awe = gate * awe
            h = awe * h

            predictions_bboxes.append(self._bbox_embed(h).sigmoid())
            predictions_classes.append(self._class_embed(h))
        if len(predictions_bboxes) > 0:
            predictions_bboxes = torch.stack([x[0] for x in predictions_bboxes])
        if len(predictions_classes) > 0:
            predictions_classes = torch.stack([x[0] for x in predictions_classes])

        return predictions_classes, predictions_bboxes



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
from __future__ import division

import collections
import numbers
import random

import torch

from docling_ibm_models.tableformer.data_management import functional as F


def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h), (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)


def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2, (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)


class Lambda(object):
    """Apply a user-defined lambda as a transform.
    Attention: The multiprocessing used in dataloader of pytorch
    is not friendly with lambda function in Windows
    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        # assert isinstance(lambd, types.LambdaType)
        self.lambd = lambd
        # if 'Windows' in platform.system():
        #     raise RuntimeError("Can't pickle lambda funciton in windows system")

    def __call__(self, img):
        return self.lambd(img)

    def __repr__(self):
        return self.__class__.__name__ + "()"


class RandomTransforms(object):
    """Base class for a list of transformations with randomness
    Args:
        transforms (list or tuple): list of transformations
    """

    def __init__(self, transforms):
        assert isinstance(transforms, (list, tuple))
        self.transforms = transforms

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


class RandomChoice(RandomTransforms):
    """Apply single transformation randomly picked from a list"""

    def __call__(self, img, target):
        t = random.choice(self.transforms)
        return t(img, target)


class RandomCrop(object):
    def __init__(self, size, margin_crop):
        self.size = list(size)
        self.margin_crop = list(margin_crop)
        # margin_crop: w, h

    def __call__(self, img, target):
        # img (w,h,ch)
        image_height, image_width = img.shape[0], img.shape[1]
        """
        img (np.ndarray): Image to be cropped.
        x: Upper pixel coordinate.
        y: Left pixel coordinate.
        h: Height of the cropped image.
        w: Width of the cropped image.
        """
        if image_width > 0 and image_height > 0:
            cropped_image = F.crop(
                img,
                self.margin_crop[1],
                self.margin_crop[0],
                image_height - (self.margin_crop[1] * 2),
                image_width - (self.margin_crop[0] * 2),
            )

            target_ = target.copy()
            target_["boxes"][:, 0] = target_["boxes"][:, 0] - self.margin_crop[0]
            target_["boxes"][:, 1] = target_["boxes"][:, 1] - self.margin_crop[1]
            target_["boxes"][:, 2] = target_["boxes"][:, 2] - self.margin_crop[0]
            target_["boxes"][:, 3] = target_["boxes"][:, 3] - self.margin_crop[1]
        else:
            cropped_image = img
        return cropped_image, target_


class RandomPad(object):
    def __init__(self, max_pad):
        self.max_pad = max_pad

    def __call__(self, img, target):
        pad_x = random.randint(0, self.max_pad)
        pad_y = random.randint(0, self.max_pad)
        pad_x1 = random.randint(0, self.max_pad)
        pad_y1 = random.randint(0, self.max_pad)
        img = img.copy()
        padded_image = F.pad(img, (pad_x, pad_y, pad_x1, pad_y1), fill=(255, 255, 255))
        target_ = target.copy()
        if target["boxes"] is not None:
            target_["boxes"][:, 0] = target_["boxes"][:, 0] + pad_x
            target_["boxes"][:, 1] = target_["boxes"][:, 1] + pad_y
            target_["boxes"][:, 2] = target_["boxes"][:, 2] + pad_x
            target_["boxes"][:, 3] = target_["boxes"][:, 3] + pad_y
        return padded_image, target_


class ColorJitter(object):
    """Randomly change the brightness, contrast and saturation of an image.
    Args:
        brightness (float): How much to jitter brightness. brightness_factor
            is chosen uniformly from [max(0, 1 - brightness), 1 + brightness].
        contrast (float): How much to jitter contrast. contrast_factor
            is chosen uniformly from [max(0, 1 - contrast), 1 + contrast].
        saturation (float): How much to jitter saturation. saturation_factor
            is chosen uniformly from [max(0, 1 - saturation), 1 + saturation].
        hue(float): How much to jitter hue. hue_factor is chosen uniformly from
            [-hue, hue]. Should be >=0 and <= 0.5.
    """

    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):

        assert isinstance(brightness, float) or (
            isinstance(brightness, collections.Iterable) and len(brightness) == 2
        )
        assert isinstance(contrast, float) or (
            isinstance(contrast, collections.Iterable) and len(contrast) == 2
        )
        assert isinstance(saturation, float) or (
            isinstance(saturation, collections.Iterable) and len(saturation) == 2
        )
        assert isinstance(hue, float) or (
            isinstance(hue, collections.Iterable) and len(hue) == 2
        )

        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    @staticmethod
    def get_params(brightness, contrast, saturation, hue):
        """Get a randomized transform to be applied on image.
        Arguments are same as that of __init__.
        Returns:
            Transform which randomly adjusts brightness, contrast and
            saturation in a random order.
        """
        transforms = []

        if isinstance(brightness, numbers.Number):

            if brightness > 0:
                brightness_factor = random.uniform(
                    max(0, 1 - brightness), 1 + brightness
                )
                transforms.append(
                    Lambda(lambda img: F.adjust_brightness(img, brightness_factor))
                )

            if contrast > 0:
                contrast_factor = random.uniform(max(0, 1 - contrast), 1 + contrast)
                transforms.append(
                    Lambda(lambda img: F.adjust_contrast(img, contrast_factor))
                )

            if saturation > 0:
                saturation_factor = random.uniform(
                    max(0, 1 - saturation), 1 + saturation
                )
                transforms.append(
                    Lambda(lambda img: F.adjust_saturation(img, saturation_factor))
                )

            if hue > 0:
                hue_factor = random.uniform(-hue, hue)
                transforms.append(Lambda(lambda img: F.adjust_hue(img, hue_factor)))

        else:

            if brightness[0] > 0 and brightness[1] > 0:

                brightness_factor = random.uniform(brightness[0], brightness[1])
                transforms.append(
                    Lambda(lambda img: F.adjust_brightness(img, brightness_factor))
                )

            if contrast[0] > 0 and contrast[1] > 0:

                contrast_factor = random.uniform(contrast[0], contrast[1])
                transforms.append(
                    Lambda(lambda img: F.adjust_contrast(img, contrast_factor))
                )

            if saturation[0] > 0 and saturation[1] > 0:

                saturation_factor = random.uniform(saturation[0], saturation[1])
                transforms.append(
                    Lambda(lambda img: F.adjust_saturation(img, saturation_factor))
                )

            if hue[0] > 0 and hue[1] > 0:
                hue_factor = random.uniform(hue[0], hue[1])
                transforms.append(Lambda(lambda img: F.adjust_hue(img, hue_factor)))

        random.shuffle(transforms)
        transform = ComposeSingle(transforms)

        return transform

    def __call__(self, img, target):
        """
        Args:
            img (np.ndarray): Input image.
        Returns:
            np.ndarray: Color jittered image.
        """
        transform = self.get_params(
            self.brightness, self.contrast, self.saturation, self.hue
        )
        return transform(img), target

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        format_string += "brightness={0}".format(self.brightness)
        format_string += ", contrast={0}".format(self.contrast)
        format_string += ", saturation={0}".format(self.saturation)
        format_string += ", hue={0})".format(self.hue)
        return format_string


class Normalize(object):
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``
    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor, target=None):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized Tensor image.
        """
        return F.normalize(tensor, self.mean, self.std), target

    def __repr__(self):
        return self.__class__.__name__ + "(mean={0}, std={1})".format(
            self.mean, self.std
        )


class NoTransformation(object):
    """Do Nothing"""

    def __call__(self, img, target):
        return img, target


class Compose(object):
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img, target):
        for t in self.transforms:
            img, target = t(img, target)
        return img, target

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


class ComposeSingle(object):
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


class Resize(object):
    """Resize the input PIL Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
        (h, w), output size will be matched to this. If size is an int,
        smaller edge of the image will be matched to this number.
        i.e, if height > width, then image will be rescaled to
        (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
        ``BILINEAR``
    """

    def __init__(self, size, interpolation="BILINEAR"):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img, target=None):
        """
        Args:
        img (np.ndarray): Image to be scaled.
        Returns:
        np.ndarray: Rescaled image.
        """
        # Resize bboxes (in pixels)
        x_scale = 0
        y_scale = 0

        if img.shape[1] > 0:
            x_scale = self.size[0] / img.shape[1]
        if img.shape[0] > 0:
            y_scale = self.size[1] / img.shape[0]

        # loop over bboxes
        if target is not None:
            if target["boxes"] is not None:
                target_ = target.copy()
                target_["boxes"][:, 0] = x_scale * target_["boxes"][:, 0]
                target_["boxes"][:, 1] = y_scale * target_["boxes"][:, 1]
                target_["boxes"][:, 2] = x_scale * target_["boxes"][:, 2]
                target_["boxes"][:, 3] = y_scale * target_["boxes"][:, 3]
        return F.resize(img, self.size, self.interpolation), target

    def __repr__(self):
        interpolate_str = self.interpolation
        return self.__class__.__name__ + "(size={0}, interpolation={1})".format(
            self.size, interpolate_str
        )
